"""Handler for selecting and running editor features"""

import sys
from . import (
    helper,
    patcher,
    serialise_save,
    adb_handler,
    server_handler,
    user_input_handler,
)

from .edits import (
    basic_items,
    other,
    gamototo,
    cats,
    levels,
)

path = ""


def save_and_exit(save_stats: dict) -> dict:
    """Serialise the save data and exit"""

    save_data = serialise_save.start_serialize(save_stats)
    helper.write_save_data(save_data, save_stats["version"], path, True)
    sys.exit()


def save_and_upload(save_stats: dict) -> dict:
    """Serialise the save data, and upload it to the game server"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = helper.write_save_data(save_data, save_stats["version"], path, False)
    upload_data = server_handler.upload_handler(save_stats, path)
    if not upload_data:
        sys.exit()
    helper.colored_text(f"Transfer code : &{upload_data['transferCode']}&")
    helper.colored_text(f"Confirmation Code : &{upload_data['pin']}&")
    sys.exit()


def fix_elsewhere_old(save_stats: dict) -> dict:
    """Fix the elsewhere error using 2 save files"""

    main_token = save_stats["token"]
    main_iq = save_stats["inquiry_code"]
    input(
        "Select a save file that is currently loaded in-game that doesn't have the elsehere error and is not banned\nPress enter to continue:"
    )
    new_path = helper.select_file(
        "Select a clean save file", helper.get_save_file_filetype(), path
    )
    if not new_path:
        print("Please select a save file")
        return save_stats

    data = helper.load_save_file(new_path)
    new_stats = data["save_stats"]
    new_token = new_stats["token"]
    new_iq = new_stats["inquiry_code"]
    save_stats["token"] = new_token
    save_stats["inquiry_code"] = new_iq

    helper.colored_text(f"Replaced inquiry code: &{main_iq}& with &{new_iq}&")
    helper.colored_text(f"Replaced token: &{main_token}& with &{new_token}&")
    return save_stats


def save_and_push(save_stats: dict) -> dict:
    """Serialise the save data and and push it to the game"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    helper.write_file_bytes(path, save_data)
    adb_handler.adb_push(save_stats["version"], path, False)
    sys.exit()


def save_and_push_rerun(save_stats: dict) -> dict:
    """Serialise the save data and push it to the game and restart the game"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    helper.write_file_bytes(path, save_data)
    adb_handler.adb_push(save_stats["version"], path, True)
    sys.exit()


def export(save_stats: dict) -> dict:
    """Export the save stats to a json file"""

    serialise_save.export_json(save_stats, path + ".json")
    sys.exit()


def clear_data(save_stats: dict) -> dict:
    """Clear data wrapper for the clear_data function"""

    confirm = input("Do want to clear your data (y/n)?:").lower()
    if confirm == "y":
        adb_handler.adb_clear(save_stats["version"])
        print("Data cleared")
    sys.exit()


FEATURES = {
    "Save Management": {
        "Save changes and upload to game servers (get transfer and confirmation codes)": save_and_upload,
        "Save changes and exit": save_and_exit,
        "Save changes and push save data to the game (don't re-open game)": save_and_push,
        "Save changes and push save data to the game (re-open game)": save_and_push_rerun,
        "Export save data as json (not desinged to be that readable)": export,
        "Clear save data (used to generate a new account without re-installing the game)": clear_data,
    },
    "Items": {
        "Cat Food": basic_items.basic.edit_cat_food,
        "XP": basic_items.basic.edit_xp,
        "Normal Tickets": basic_items.basic.edit_normal_tickets,
        "Rare Tickets": basic_items.basic.edit_rare_tickets,
        "Platinum Tickets": basic_items.basic.edit_platinum_tickets,
        "Platinum Shards": basic_items.basic.edit_platinum_shards,
        "Legend Tickets": basic_items.basic.edit_legend_tickets,
        "NP": basic_items.basic.edit_np,
        "Leadership": basic_items.basic.edit_leadership,
        "Battle Items": basic_items.basic.edit_battle_items,
        "Catseyes": basic_items.basic.edit_catseyes,
        "Cat Fruit / Behemoth Stones": basic_items.basic.edit_catfruit,
        "Talent Orbs": basic_items.talent_orbs.edit_talent_orbs,
    },
    "Gamatoto / Ototo": {
        "Ototo Engineers": basic_items.basic.edit_engineers,
        "Base materials": basic_items.basic.edit_base_materials,
        "Catamins": basic_items.basic.edit_catamins,
        "Gamatoto XP / Level": gamototo.gamatoto_xp.edit_gamatoto_xp,
        "Ototo Cat Cannon": gamototo.ototo_cat_cannon.edit_cat_cannon,
        "Gamatoto Helpers": gamototo.helpers.edit_helpers,
        "Fix gamatoto from crashing the game": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "Cats / Special Skills": {
        "Get Cats": cats.get_remove_cats.get_cat,
        "Get Cats Based On Rarity": cats.get_remove_cats.get_cat_rarity,
        "Remove Cats": cats.get_remove_cats.remove_cats,
        "Upgrade Cats": cats.upgrade_cats.upgrade_cats,
        "Upgrade Currently Unlocked Cats": cats.upgrade_cats.upgrade_current_cats,
        "Upgrade Cats Based On Rarity": cats.upgrade_cats.upgrade_cat_rarity,
        "Upgrade Special Skills / Base upgrades (The ones that are blue)": cats.upgrade_blue.upgrade_blue,
        "True Form Currently Unlocked Cats": cats.evolve_cats.get_evolve_current,
        "True Form Cats": cats.evolve_cats.get_evolve,
        "True Form Cats Based On Rarity": cats.evolve_cats.evolve_cat_rarity,
        "Force True Form Cats": cats.evolve_cats.get_evolve_forced,
        "Remove True Forms": cats.evolve_cats.remove_evolve,
        "Talents": cats.talents.edit_talents,
        "Collect Cat Guide": cats.clear_cat_guide.clear_cat_guide,
        "Collect Cat Guide Based On Rarity": cats.clear_cat_guide.clear_cat_guide_rarity,
        'Get stage unit drops - removes the "Clear this stage to get special cat" dialog': cats.chara_drop.get_character_drops,
    },
    "Levels / Treasures": {
        "Main Story Chapters Clear": levels.main_story.main_story,
        "Whole Chapter Main Story Treasures": levels.treasures.treasures,
        "Individual Treasures Per Stage / Treasure Groups (e.g energy drink, aqua crystal)": levels.treasures.specific_treasures,
        "Zombie Stages / Outbreaks": levels.outbreaks.edit_outbreaks,
        "Event Stages": levels.event_stages.event_stages,
        "Stories of Legend": levels.event_stages.stories_of_legend,
        "Uncanny Legends": levels.uncanny.edit_uncanny,
        "Aku Realm/Gates": levels.aku.edit_aku,
        "Gauntlets": levels.gauntlet.edit_gauntlet,
        "Into the Future Timed Scores": levels.itf_timed_scores.timed_scores,
        "Challenge Battle Score": basic_items.basic.edit_challenge_battle,
        "Towers": levels.towers.edit_tower,
        "Clear Tutorial": levels.clear_tutorial.clear_tutorial,
        "Catclaw Dojo Score (Hall of Initiates)": basic_items.basic.edit_dojo_score,
    },
    "Inquiry Code / Token": {
        "Inquiry Code": basic_items.basic.edit_inquiry_code,
        "Token": basic_items.basic.edit_token,
        "Fix elsewhere error / Unban account": other.fix_elsewhere.fix_elsewhere,
        "Old Fix elsewhere error / Unban account (needs 2 save files)": fix_elsewhere_old,
    },
    "Other": {
        "Rare Gacha Seed": basic_items.basic.edit_rare_gacha_seed,
        "Unlocked Equip Slots": basic_items.basic.edit_unlocked_slots,
        "Restart Pack": basic_items.basic.edit_restart_pack,
        "Meow Medals": other.meow_medals.medals,
        "Play Time": other.play_time.edit_play_time,
        "Unlock / Remove Enemy Guide Entries": other.unlock_enemy_guide.enemy_guide,
        "Catnip Challenges / Missions": other.missions.edit_missions,
        "Normal Ticket Max Trade Progress (allows for unbannable rare tickets)": other.trade_progress.set_trade_progress,
    },
}


def get_feature(features: dict, search_string: str, results: dict) -> dict:
    """Search for a feature if the feature name contains the search string"""

    for feature in features:
        feature_data = features[feature]
        if isinstance(feature_data, dict):
            feature_data = get_feature(feature_data, search_string, results)
        if search_string.lower().replace(" ", "") in feature.lower().replace(" ", ""):
            results[feature] = features[feature]
    return results


def show_options(save_stats: dict, features: dict) -> dict:
    """Allow the user to either enter a feature number or a feature name, and get the features that match"""

    user_input = user_input_handler.colored_input(
        "What do you want to edit (some options contain other features within them)\nYou can enter a number to run a feature or a word to search for that feature (e.g entering catfood will run the Cat Food feature, and entering tickets will show you all the features that edit tickets)\nYou can press enter to see a list of all of the features:\n"
    )
    user_int = helper.check_int(user_input)
    results = []
    if user_int is None:
        results = get_feature(features, user_input, {})
    else:
        if user_int < 1 or user_int > len(features) + 1:
            helper.colored_text("Value out of range", helper.RED)
            return show_options(save_stats, features)
        if FEATURES != features:
            if user_int - 2 < 0:
                return menu(save_stats)
            results = features[list(features)[user_int - 2]]
        else:
            results = features[list(features)[user_int - 1]]
    if not isinstance(results, dict):
        return results(save_stats)
    if len(results) == 0:
        helper.colored_text("No feature found with that name.", helper.RED)
        return menu(save_stats)
    if len(results) == 1:
        return results[list(results)[0]](save_stats)

    helper.colored_list(["Go Back"] + list(results))
    return show_options(save_stats, results)


def menu(save_stats: dict, path_save: str = None) -> dict:
    """Show the menu and allow the user to select a feature to edit"""

    if path_save:
        global path
        path = path_save
    helper.colored_list(list(FEATURES))
    save_stats = show_options(save_stats, FEATURES)

    return save_stats



"""Handler for clearing gauntlets"""

from ..levels import event_stages
from ... import user_input_handler


def edit_gauntlet(save_stats: dict) -> dict:
    """Handler for clearing gauntlets"""

    stage_data = save_stats["gauntlets"]
    lengths = stage_data["Lengths"]

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter gauntlet ids (Look up &Event Release Order battle cats& and scroll past the &events& to find &gauntlet& ids) (You can enter &all& to get all, a range e.g 1-49, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths["total"],
    )
    save_stats["gauntlets"] = event_stages.stage_handler(stage_data, ids, 0)
    return save_stats



"""Handler to add and remove cats"""

from ..cats import upgrade_cats
from ... import user_input_handler, helper


def get_cat(save_stats: dict) -> dict:
    """Handler to get cats"""

    return cat_handler(save_stats, 1, "gave")


def remove_cats(save_stats: dict) -> dict:
    """Handler to remove cats"""

    return cat_handler(save_stats, 0, "removed")


def get_cat_rarity(save_stats: dict) -> dict:
    """Get all cats of a certain rarity"""

    ids = user_input_handler.select_options(
        options=upgrade_cats.types,
        mode="get",
    )
    cat_ids = upgrade_cats.get_rarity(ids)

    save_stats = get_cat_ids(
        save_stats=save_stats,
        val=1,
        string="gave",
        ids=cat_ids,
    )
    return save_stats


def cat_handler(save_stats: dict, val: int, string: str) -> dict:
    """Get specific cats"""

    cats = save_stats["cats"]
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        len(cats),
    )

    save_stats = get_cat_ids(
        save_stats=save_stats,
        val=val,
        string=string,
        ids=ids,
    )

    return save_stats


def get_cat_ids(save_stats: dict, val: int, string: str, ids: list) -> dict:
    """Get specific cats by ids"""

    ids = helper.check_cat_ids(ids, save_stats)

    cats = save_stats["cats"]

    for cat_id in ids:
        cats[cat_id] = val

    save_stats["cats"] = cats
    print(f"Successfully {string} cats")
    return save_stats



"""Handler for editing gamatoto helpers"""

from ... import helper as helper_module
from ... import item


def get_gamatoto_helpers() -> dict:
    """Get the rarities of all gamatoto helpers"""

    data = helper_module.read_file_string(
        helper_module.get_file(
            "game_data/gamatoto/GamatotoExpedition_Members_name_en.csv"
        )
    ).splitlines()
    helpers = {}
    for line in data:
        line_data = line.split("|")
        if len(line_data) < 5:
            break

        helper_id = line_data[0]
        rarity = int(line_data[1])
        type_str = line_data[4]
        helpers[helper_id] = {"Rarity_id": rarity, "Rarity_name": type_str}
    return helpers


def generate_helpers(user_input: list, helper_data: dict) -> list:
    """Generate unique helpers from amounts of each"""

    final_helpers = []
    values = list(helper_data.values())
    for i, usr_input in enumerate(user_input):
        for j, value in enumerate(values):
            if value["Rarity_id"] == i:
                final_helpers += list(range(j + 1, j + 1 + usr_input))
                break
    return final_helpers


def edit_helpers(save_stats: dict) -> dict:
    """Handler for gamatoto helpers"""

    helpers = save_stats["helpers"]

    helper_data = get_gamatoto_helpers()
    current_helpers = {}
    helper_count = {
        "Intern": 0,
        "Lackey": 0,
        "Underling": 0,
        "Assistant": 0,
        "Legend": 0,
    }
    for helper in helpers:
        if helper == 0xFFFFFFFF:
            break
        current_helpers[helper] = helper_data[str(helper)]
        helper_count[current_helpers[helper]["Rarity_name"]] += 1

    helpers_counts_input = item.create_item_group(
        names=list(helper_count.keys()),
        values=list(helper_count.values()),
        edit_name="amount",
        group_name="Gamatoto Helpers",
        maxes=10,
    )
    helpers_counts_input.edit()
    final_helpers = generate_helpers(helpers_counts_input.values, helper_data)
    extra_ls = [0xFFFFFFFF] * (len(helpers) - len(final_helpers))
    final_helpers += extra_ls

    helpers = final_helpers
    save_stats["helpers"] = helpers
    return save_stats



"""Handler for clearing event stages"""
from ... import helper, user_input_handler


def stage_handler(
    stage_data: dict, ids: list, offset: int, unlock_next: bool = True
) -> dict:
    """Clear stages from a set of ids"""

    lengths = stage_data["Lengths"]

    individual = True
    if len(ids) > 1:
        individual = user_input_handler.colored_input(
            "Do you want to set the stars/crowns for each subchapter individually(&1&), or all at once(&2&):"
        ) == "1"
    first = True
    stars = None
    for stage_id in ids:
        if not individual and first:
            stars = helper.check_int(
                user_input_handler.colored_input(
                    f"Enter the number of stars/cowns (max &{lengths['stars']}&):"
                )
            )
            if stars is None:
                print("Please enter a valid number")
                continue
            stars = helper.clamp(stars, 0, lengths["stars"])
            first = False
        elif individual:
            stars = helper.check_int(
                user_input_handler.colored_input(
                    f"Enter the number of stars/cowns for subchapter &{stage_id}& (max &{lengths['stars']}&):"
                )
            )
            if stars is None:
                print("Please enter a valid number")
                continue
            stars = helper.clamp(stars, 0, lengths["stars"])
        stage_id += offset
        stage_data_edit = stage_data
        if stage_id >= len(stage_data_edit["Value"]["clear_progress"]):
            continue
        stage_data_edit["Value"]["clear_progress"][stage_id] = (
            [lengths["stages"]] * stars
        ) + ([0] * (lengths["stars"] - stars))
        if unlock_next and stage_id + 1 < len(
            stage_data_edit["Value"]["clear_progress"]
        ):
            stage_data_edit["Value"]["unlock_next"][stage_id + 1] = (
                [lengths["stars"] - 1] * stars
            ) + ([0] * (lengths["stars"] - stars))
        stage_data_edit["Value"]["clear_amount"][stage_id] = (
            [[1] * lengths["stages"]] * stars
        ) + ([[0] * lengths["stages"]] * (lengths["stars"] - stars))

    print("Successfully set subchapters")

    return stage_data_edit


def stories_of_legend(save_stats: dict) -> dict:
    """Handler for clearing stories of legend"""

    stage_data = save_stats["event_stages"]

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter subchapter ids (e.g &1& = legend begins, &2& = passion land)(You can enter &all& to get all, a range e.g &1&-&49&, or ids separate by spaces e.g &5 4 7&):"
        ),
        50,
    )
    offset = -1
    save_stats["event_stages"] = stage_handler(stage_data, ids, offset)
    return save_stats


def event_stages(save_stats: dict) -> dict:
    """Handler for clearing event stages"""

    stage_data = save_stats["event_stages"]
    lengths = stage_data["Lengths"]

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter subchapter ids (Look up &Event Release Order battle cats& to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        lengths["total"] - 400,
    )
    offset = 400
    save_stats["event_stages"] = stage_handler(stage_data, ids, offset)
    return save_stats



"""Handler for editting outbreaks"""

from ... import helper, user_input_handler
from ..levels import main_story


def encode_ls(lst: list) -> dict:
    """Encode a list of integers into a dictionary"""
    return {i: lst[i] for i in range(len(lst))}


def edit_outbreaks(save_stats: dict) -> dict:
    """Handler for editting outbreaks"""

    outbreaks = save_stats["outbreaks"]["outbreaks"]

    available_chapters = []
    for chapter in outbreaks.keys():
        index = chapter
        if index > 2:
            index -= 1
        available_chapters.append(main_story.chapters[index])

    print("What chapter do you want to edit:")
    ids = user_input_handler.select_options(
        options=available_chapters,
        mode="clear the outbreaks for?",
    )
    ids = helper.check_clamp(ids, len(available_chapters), 1, -1)
    ids = main_story.format_story_ids(ids)
    for chapter_id in ids:
        outbreaks[chapter_id] = encode_ls([1] * len(outbreaks[chapter_id]))
    save_stats["outbreaks"]["outbreaks"] = outbreaks
    print("Successfully set outbreaks")
    return save_stats



"""Handler for editting meow medals"""

from ... import helper, user_input_handler


def get_medal_names() -> list:
    """Get all medal names"""

    medal_names = helper.read_file_string(
        helper.get_file("game_data/medals/medalname.tsv")
    ).splitlines()
    names = []
    for line in medal_names:
        line_split = line.split("\t")
        name = (
            line_split[0]
            .rstrip("\n")
            .replace("&", "and")
            .replace("â˜…", "")
            .lstrip(" ")
        )
        names.append(name)
    return names


def medals(save_stats: dict) -> dict:
    """Handler for editting meow medals"""

    medal_stats = save_stats["medals"]

    names = get_medal_names()
    helper.colored_list(names)

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter medal ids (You can enter all to get &all&, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        len(names) + 1,
    )

    for medal_id in ids:
        if medal_id == 0:
            continue
        medal_id -= 1
        if medal_id not in medal_stats["medal_data_1"]:
            if medal_id not in medal_stats["medal_data_2"]:
                medal_stats["medal_data_1"].append(medal_id)
            medal_stats["medal_data_2"][medal_id] = 0
    save_stats["medals"] = medal_stats
    print("Successfully gave medals")
    return save_stats



"""Handler for editting trade progress to allow for unbannable rare tickets"""

from ... import helper, item


def set_trade_progress(save_stats: dict) -> dict:
    """Handler for editting trade progress to allow for unbannable rare tickets"""

    trade_progress = save_stats["trade_progress"]
    storage = save_stats["cat_storage"]
    tickets = item.Item(
        name="Rare Tickets",
        max_value=299,
        value=save_stats["rare_tickets"]["Value"],
        edit_name="amount",
        set_name="gain",
    )
    tickets.edit()
    trade_progress["Value"] = tickets.value * 5

    space = False

    for i in range(len(storage["types"])):
        storage_item = storage["types"][i]
        if storage_item == 0:
            storage["ids"][i] = 1
            storage["types"][i] = 2
            space = True
            break
    if not space:
        helper.colored_text("Your cat storage is full, please free 1 space!")
        return save_stats

    save_stats["cat_storage"] = storage
    save_stats["trade_progress"] = trade_progress
    helper.colored_text(
        'You now need to go into your storage and press &"Use all"& and then press &"Trade for Ticket"&'
    )

    return save_stats



"""Handler to edit cat talents"""
from ... import helper, user_input_handler, csv_file_handler, item


def get_talent_data() -> dict:
    """Get talent data for all cats"""

    talent_data_raw = csv_file_handler.parse_csv(
        helper.get_file("game_data/talents/SkillAcquisition.csv")
    )
    talent_names = csv_file_handler.parse_csv(
        helper.get_file("game_data/talents/SkillDescriptions.csv"), "|", parse_int=False
    )
    columns = talent_data_raw[0]
    new_talent_data = {}
    for j in range(1, len(talent_data_raw)):
        data = talent_data_raw[j]
        cat_id = data[0]
        new_talent_data[cat_id] = {}

        for data, column in zip(data, columns):
            new_talent_data = replace_name(
                cat_id=cat_id,
                column=column,
                data=data,
                talent_names=talent_names,
                new_data=new_talent_data,
            )
    return new_talent_data


def replace_name(
    cat_id: int, column: str, data: int, talent_names: list[list[str]], new_data: dict
) -> dict:
    """Replace the text ids with the corresponding names"""

    new_data[cat_id][column] = data
    if "textID" in column:
        new_data[cat_id][column] = talent_names[data][1]
        stop_at = "<br>"
        if stop_at in new_data[cat_id][column]:
            index = new_data[cat_id][column].index(stop_at)
            new_data[cat_id][column] = new_data[cat_id][column][:index]
    return new_data


def find_order(cat_talents: dict, cat_talent_data: dict) -> list:
    """Find what talent slot each letter corresponds to"""

    letters = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
    ]
    letter_order = []

    for i, talent in enumerate(cat_talents):
        if i == 0:
            continue
        talent_id = talent["id"]
        for letter in letters:
            ability_id = int(cat_talent_data[f"abilityID_{letter}"])
            if ability_id == talent_id:
                letter_order.append(letter)
    return letter_order


def get_cat_talents(cat_talents: dict, cat_talent_data: dict) -> dict:
    """Get the name and max value of each talent for a specific cat"""

    data = {}
    letter_order = find_order(cat_talents, cat_talent_data)
    for i in range(len(cat_talents) - 1):
        cat_data = {}
        cat_data["name"] = cat_talent_data[f"textID_{letter_order[i]}"].strip("\n")
        cat_data["max"] = int(cat_talent_data[f"MAXLv_{letter_order[i]}"])
        if cat_data["max"] == 0:
            cat_data["max"] = 1
        data[i] = cat_data
    return data


def edit_talents(save_stats: dict) -> dict:
    """Handler for editing talents"""

    length = len(save_stats["cats"])
    talents = save_stats["talents"]
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        length,
    )
    individual = True
    if len(ids) > 1:
        individual = user_input_handler.colored_input(
            "Do you want to set the talents for each cat individually(&1&), or max them all at once(&2&):"
        ) == "1"

    talent_data = get_talent_data()
    cat_talents_levels = []
    for cat_id in ids:
        if cat_id not in talents or cat_id not in talent_data:
            # don't spam the user with messages if they selected alot of ids at once
            if len(ids) < 20:
                print(f"Error cat {cat_id} does not have any talents")
            continue
        if not individual:
            cat_talent_data = talent_data[cat_id]
            cat_talents = talents[cat_id]
            cat_talent_data_formatted = get_cat_talents(cat_talents, cat_talent_data)
            names = []
            maxes = []
            cat_talents_levels = []
            for talent_formatted in cat_talent_data_formatted.values():
                max_val = talent_formatted["max"]
                cat_talents_levels.append(max_val)

        elif individual:
            cat_talent_data = talent_data[cat_id]
            cat_talents = talents[cat_id]
            cat_talent_data_formatted = get_cat_talents(cat_talents, cat_talent_data)
            names = []
            cat_talents_levels = []
            maxes = []
            for talent_index, cat_talent_formatted in cat_talent_data_formatted.items():
                names.append(cat_talent_formatted["name"])
                cat_talents_levels.append(cat_talents[talent_index + 1]["level"])
                maxes.append(cat_talent_formatted["max"])
            helper.colored_text(f"Cat &{cat_id}& is slected:")
            cat_talents_levels = item.create_item_group(
                names=names,
                values=cat_talents_levels,
                maxes=maxes,
                edit_name="talents",
                group_name="Talents"
            )
            cat_talents_levels.edit()
            cat_talents_levels = cat_talents_levels.values

        for i, cat_talent_level in enumerate(cat_talents_levels):
            cat_talents[i + 1]["level"] = cat_talent_level
        save_stats["talents"] = talents

    print("Successfully set talents")
    return save_stats



"""Handler for editing character drops"""

from ... import helper, csv_file_handler, user_input_handler

def get_data() -> dict:
    """gets all of the cat ids and treasure ids that can be dropped"""

    character_data = csv_file_handler.parse_csv(helper.get_file("game_data/drops/drop_chara.csv"))[1:]

    treasure_ids = helper.copy_first_n(character_data, 0)
    indexes = helper.copy_first_n(character_data, 1)
    cat_ids = helper.copy_first_n(character_data, 2)

    return {"t_ids": treasure_ids, "indexes": indexes, "c_ids": cat_ids}


def set_t_ids(save_stats: dict) -> dict:
    """handler for editing treasure ids"""

    unit_drops_stats = save_stats["unit_drops"]
    data = get_data()

    usr_t_ids = user_input_handler.get_range(user_input_handler.colored_input("Enter treasures ids (Look up item drop cats battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"), all_ids=data["t_ids"])

    for t_id in usr_t_ids:
        if t_id in data["t_ids"]:
            index = data["t_ids"].index(t_id)
            save_index = data["indexes"][index]
            unit_drops_stats[save_index] = 1

    save_stats["unit_drops"] = unit_drops_stats
    return save_stats

def set_c_ids(save_stats: dict) -> dict:
    """handler for editing cat ids"""

    unit_drops_stats = save_stats["unit_drops"]
    data = get_data()

    usr_c_ids = user_input_handler.get_range(user_input_handler.colored_input("Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"), all_ids=data["c_ids"])
    usr_c_ids = helper.check_cat_ids(usr_c_ids, save_stats)
    for c_id in usr_c_ids:
        if c_id in data["c_ids"]:
            index = data["c_ids"].index(c_id)
            save_index = data["indexes"][index]
            unit_drops_stats[save_index] = 1

    save_stats["unit_drops"] = unit_drops_stats
    return save_stats

def get_character_drops(save_stats: dict) -> dict:
    """handler for getting character drops"""

    flag_t_ids = user_input_handler.colored_input("Do you want to select treasure ids &(1)&, or cat ids? &(2)&:") == "1"

    if flag_t_ids:
        save_stats = set_t_ids(save_stats)
    else:
        save_stats = set_c_ids(save_stats)
    print("Successfully set unit drops")

    return save_stats



"""Handler for unlocking the enemy guide"""

from ... import user_input_handler


def enemy_guide(save_stats: dict) -> dict:
    """Handler for unlocking the enemy guide"""

    enemy_guide_stats = save_stats["enemy_guide"]
    total = len(enemy_guide_stats)
    unlock = user_input_handler.colored_input(
        "Do you want to remove enemy guide entries &(1)& or unlock them &(2)&:"
    ) == "2"
    set_val = 1
    if not unlock:
        set_val = 0
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter enemy ids (Look up enemy release order battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        total,
    )

    for enemy_id in ids:
        if enemy_id >= 2:
            enemy_id -= 2
        if enemy_id >= len(enemy_guide_stats):
            print(f"Invalid enemy id: {enemy_id+2}")
            continue
        enemy_guide_stats[enemy_id] = set_val
    save_stats["enemy_guide"] = enemy_guide_stats
    if not unlock:
        print("Successfully removed enemy guide entries")
    else:
        print("Successfully unlocked enemy guide entries")
    return save_stats



"""Module that runs when the module is run directly"""

import sys
from . import (
    helper,
    adb_handler,
    server_handler,
    updater,
    patcher,
    feature_handler,
    serialise_save,
    parse_save,
)


def main():
    """Main function"""

    helper.colored_text(
        "Welcome to the Battle Cats Save File Editor\nMade by fieryhenry\nhttps://github.com/fieryhenry/BCSFE-Python\n",
        base=helper.CYAN,
    )

    updater.check_update()
    normal_start_up()


def normal_start_up() -> None:
    """Display and handle options for downloading save data, pulling save data, selecting save data"""

    print()
    options = [
        "Download save data from the game using transfer and confirmation codes",
        "Select a save file from file",
        "Use adb to pull the save from a rooted device",
        "Load save data from json",
    ]
    helper.colored_list(options)
    option = input(f"Enter an option (1 to {len(options)}):")
    path = None
    if option == "1":
        path = server_handler.download_save_handler()
    elif option == "2":
        path = helper.select_file(
            "Select a save file:", helper.get_save_file_filetype()
        )
    elif option == "3":
        path = adb_handler.adb_pull_handler()
    elif option == "4":
        print("Select save data json file")
        js_path = helper.select_file("Select save data json file", [("Json", "*.json")])
        if js_path:
            path = helper.load_json_handler(js_path)
    else:
        print("Please enter a recognised option:")
        return normal_start_up()
    if not path:
        return normal_start_up()
    start(path)
    return None


def start(path: str) -> None:
    """Parse, patch, start the editor and serialise the save data"""
    data = helper.load_save_file(path)
    save_stats = data["save_stats"]
    save_data = data["save_data"]
    country_code = data["country_code"]

    if path.endswith(".json"):
        input(
            "Your save data seems to be in json format. Please use to import json option if you want to load json data.\nPress enter to continue...:"
        )
    while True:
        save_stats = parse_save.start_parse(save_data, country_code)
        save_data = patcher.patch_save_data(save_data, country_code)
        save_stats = feature_handler.menu(save_stats, path)
        save_data = serialise_save.start_serialize(save_stats)
        save_data = patcher.patch_save_data(save_data, country_code)


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit()



"""Handler for clearing the cat guide"""

from ..cats import upgrade_cats
from ... import user_input_handler, helper


def clear_cat_guide(save_stats: dict) -> dict:
    """clear cat guide for specific cats"""

    cat_guide_collected = save_stats["cat_guide_collected"]

    total = len(cat_guide_collected)
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        total,
    )
    save_stats = cat_guide_ids(save_stats, ids)
    return save_stats


def clear_cat_guide_rarity(save_stats: dict) -> dict:
    """Clear cat guide for all cats of a certain rarity"""

    ids = user_input_handler.select_options(
        options=upgrade_cats.types,
        mode="collect",
    )
    cat_ids = upgrade_cats.get_rarity(ids)
    save_stats = cat_guide_ids(save_stats, cat_ids)
    return save_stats


def cat_guide_ids(save_stats: dict, ids: list) -> dict:
    """Clear cat guide for a set of cat ids"""
    ids = helper.check_cat_ids(ids, save_stats)
    cat_guide_collected = save_stats["cat_guide_collected"]
    for cat_id in ids:
        cat_guide_collected[cat_id] = 1

    save_stats["cat_guide_collected"] = cat_guide_collected
    print("Successfully collected cat guide")
    return save_stats



"""Handler for editing the aku realm"""


def edit_aku(save_stats: dict) -> dict:
    """Handler for editing the aku realm"""
    aku = save_stats["aku"]
    count = aku["Lengths"]["stages"]
    aku["Value"]["clear_amount"][0][0] = [1] * count
    save_stats["aku"] = aku
    print("Successfully cleared the aku realm")
    return save_stats



"""Handler for editing catnip missions"""
from ... import helper, user_input_handler, csv_file_handler


def get_mission_conditions() -> dict:
    """Get the mission data and what you need to do to complete it"""

    mission_condition_path = helper.get_file("game_data/missions/Mission_Condition.csv")
    mission_conditions_list = csv_file_handler.parse_csv(mission_condition_path)
    mission_conditions = {}
    for line in mission_conditions_list[1:]:
        mission_id = line[0]
        mission_conditions[mission_id] = {
            "mission_type": line[1],
            "conditions_type": line[2],
            "progress_count": line[3],
            "conditions_value": line[4:],
        }
    return mission_conditions


def get_mission_names() -> dict:
    """Get all mission names"""

    mission_name = helper.read_file_string(
        helper.get_file("game_data/missions/Mission_Name.csv")
    )
    mission_name_list = mission_name.split("\n")
    mission_names = {}
    for mission_name in mission_name_list:
        line_data = mission_name.split("|")
        mission_id = int(line_data[0])
        mission_names[mission_id] = line_data[1]
    return mission_names


def get_mission_names_from_ids(ids: list, mission_names: dict) -> list:
    """Get the mission names from the ids"""

    names = []
    for mission_id in ids:
        if mission_id in mission_names:
            names.append(mission_names[mission_id])
    return names


def edit_missions(save_stats: dict) -> dict:
    """Handler for editting catnip missions"""

    missions = save_stats["missions"]

    names = get_mission_names()
    conditions = get_mission_conditions()

    mission_ids_to_use = []
    for mission_id in missions["flags"]:
        if mission_id in conditions:
            mission_ids_to_use.append(mission_id)

    names_to_use = get_mission_names_from_ids(mission_ids_to_use, names)

    ids = user_input_handler.select_options(
        options=names_to_use,
        mode="complete",
    )
    re_claim = (
        user_input_handler.colored_input(
            "Do you want to re-complete already claimed missions &(1)& (Allows you to get the rewards again) or only complete non-claimed missions&(2)&:"
        )
        == "1"
    )
    for mission_id in ids:
        mission_id = helper.clamp(mission_id, 1, len(mission_ids_to_use))
        mission_id -= 1
        mission_id = mission_ids_to_use[mission_id]
        if re_claim:
            claim = True
        elif not re_claim and missions["flags"][mission_id] != 4:
            claim = True
        else:
            claim = False
        if claim:
            missions["flags"][mission_id] = 2
            missions["values"][mission_id] = conditions[mission_id]["progress_count"]

    save_stats["missions"] = missions
    print("Successfully completed missions")
    return save_stats



"""Handler for user input"""

from typing import Tuple, Union

from . import helper


def handle_all_at_once(
    ids: list,
    all_at_once: bool,
    data: list,
    names: list,
    item_name: str,
    group_name: str,
    explain_text: str = "",
) -> list:
    """Handle all at once option"""

    first = True
    value = 0
    for item_id in ids:
        if all_at_once and first:
            value = helper.check_int(
                colored_input(f"Enter {item_name} {explain_text}:")
            )
            first = False
        elif not all_at_once:
            value = helper.check_int(
                colored_input(
                    f"Enter {item_name} for {group_name} &{names[item_id]}& {explain_text}:"
                )
            )
        if value is None:
            continue
        data[item_id] = value
    return data


def create_all_list(
    ids: list, max_val: int, include_all_at_once: bool = False
) -> Union[dict, list]:
    """Creates a list with an all at once option"""

    all_at_once = False
    if f"{max_val}" in ids:
        ids = range(1, max_val)
        ids = [format(x, "02d") for x in ids]
        all_at_once = True
    if include_all_at_once:
        return {"ids": ids, "at_once": all_at_once}
    return ids


def get_range(
    usr_input: str, length: int = None, min_val: int = 0, all_ids: list = None,
) -> list:
    """Get a range of numbers from user input"""

    ids = []
    if usr_input.lower() == "all":
        if length is None and all_ids is None:
            helper.colored_text("You can't use &all& here", helper.RED)
            return []
        if all_ids:
            return all_ids
        return range(min_val, length)
    if "-" in usr_input:
        start, end = usr_input.split("-")
        start = helper.check_int(start)
        end = helper.check_int(end)
        if start is None or end is None:
            helper.colored_text(
                "Invalid input. Please enter a valid range of numbers separated by a dash.",
                helper.RED,
            )
            return ids
        if start > end:
            start, end = end, start
        ids = range(start, end + 1)
    else:
        content = usr_input.split(" ")
        for item_id in content:
            item_id = helper.check_int(item_id)
            if item_id is None:
                helper.colored_text(
                    "Invalid input. Please enter a valid integer.", helper.RED
                )
                return ids
            ids.append(item_id)
    return ids


def colored_input(dialog: str) -> str:
    """Format dialog as a colored string"""

    helper.colored_text(dialog, end="")
    return input()


def get_range_ids(group_name: str, length: int) -> list[int]:
    """Get a range of ids from user input"""

    ids = get_range(
        colored_input(
            f"Enter {group_name} ids(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        length,
    )
    return ids


def select_options(
    options: list,
    mode: str = "edit",
    extra_data: list = None,
    include: bool = False,
    offset: int = 0,
) -> Union[Tuple[list, bool], list]:
    """Select an option or multiple options from a list"""

    helper.colored_list(options, extra_data=extra_data, offset=offset)
    total = len(options)
    helper.colored_text(f"{total+1}. &Select all&")
    ids = colored_input(
        f"What do you want to {mode} (You can enter multiple values separated by spaces to {mode} multiple at once):"
    ).split(" ")
    individual = True
    if str(total + 1) in ids:
        ids = range(1, total + 1)
        individual = False
        ids = helper.int_to_str_ls(ids)
    ids = helper.parse_int_list(ids, -1)
    if include:
        return ids, individual
    return ids


def get_int(dialog: str) -> int:
    """Get user input as an integer and keep asking until a valid integer is entered"""

    helper.colored_text(dialog, end="")
    while True:
        try:
            val = input()
            val = val.strip(" ")
            return int(val)
        except ValueError:
            helper.colored_text(
                "Invalid input. Please enter a valid integer.", helper.RED
            )



"""Handler for setting into the future timed scores"""

from ..levels import main_story
from ... import item


def timed_scores(save_stats: dict) -> dict:
    """Handler for setting into the future timed scores"""

    scores = save_stats["itf_timed_scores"]

    usr_scores = item.create_item_group(
        names=main_story.chapters[3:6],
        values=None,
        maxes=9999,
        edit_name="score",
        group_name="Into The Future Timed Scores",
    )
    usr_scores.edit()
    for i, usr_score in enumerate(usr_scores.values):
        if usr_score is not None:
            scores[i] = ([usr_score] * 48) + ([0] * 3)
    save_stats["itf_timed_scores"] = scores

    print("Successfully set timed scores")
    return save_stats



"""Fix gamatoto from crashing the game"""


def fix_gamatoto(save_stats: dict) -> dict:
    """Fix gamatoto from crashing the game"""

    save_stats["gamatoto_skin"]["Value"] = 2
    print("Successfully fixed gamatoto from crashing the game")
    return save_stats



"""Handler for evolving cats"""

from ... import helper, user_input_handler, csv_file_handler
from ..cats import upgrade_cats


def get_evolve(save_stats: dict) -> dict:
    """Handler for evolving cats"""

    return evolve_handler(save_stats, 2, "set", False)


def get_evolve_forced(save_stats: dict) -> dict:
    """Handler for evolving cats without the form check"""

    return evolve_handler(save_stats, 2, "set", True)


def remove_evolve(save_stats: dict) -> dict:
    """Handler for de-evolving cats"""

    return evolve_handler(save_stats, 0, "removed", True)


def get_evolve_current(save_stats: dict) -> dict:
    """Evolve current cats"""

    cats = save_stats["cats"]
    current_cats = []
    for i, cat_flag in enumerate(cats):
        if cat_flag == 1:
            current_cats.append(i)

    return evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=current_cats,
        forced=False,
    )


def evolve_cat_rarity(save_stats: dict) -> dict:
    """Evolve all cats of a certain rarity"""

    ids = user_input_handler.select_options(
        options=upgrade_cats.types,
        mode="true form",
    )
    cat_ids = upgrade_cats.get_rarity(ids)

    save_stats = evolve_handler_ids(
        save_stats=save_stats,
        val=2,
        string="set",
        ids=cat_ids,
        forced=False,
    )
    return save_stats


def get_evolve_data() -> list:
    """Get max form of cats"""

    data = csv_file_handler.parse_csv(
        helper.get_file("game_data/true_forms/nyankoPictureBookData.csv")
    )
    forms = helper.copy_first_n(data, 2)
    forms = helper.offset_list(forms, -1)
    return forms


def evolve_handler(save_stats: dict, val: int, string: str, forced: bool) -> dict:
    """Evolve specific cats"""

    evolves = save_stats["unlocked_forms"]
    flags = evolves
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        len(flags),
    )
    return evolve_handler_ids(save_stats, val, string, ids, forced)


def evolve_handler_ids(
    save_stats: dict, val: int, string: str, ids: list, forced: bool
) -> dict:
    """Evolve specific cats by ids"""
    ids = helper.check_cat_ids(ids, save_stats)
    evolves = save_stats["unlocked_forms"]
    if not forced:
        form_data = get_evolve_data()
        length = min([len(ids), len(form_data)])
        for i in range(length):
            evolves[ids[i]] = form_data[i]
    else:
        for cat_id in ids:
            evolves[cat_id] = val
    save_stats["current_forms"] = evolves

    flags_evolved = [0 if form == 1 else form for form in evolves]
    save_stats["unlocked_forms"] = flags_evolved

    print(f"Successfully {string} true forms of cats")
    return save_stats



"""Handler for server requests"""

import hashlib
import hmac
import random
import json
import time
from typing import Union
import requests
from . import serialise_save, helper, patcher


def get_current_time_stamp() -> int:
    """Returns the current time stamp in seconds"""

    return int(time.time())


def random_hex(length: int) -> str:
    """Returns a random hex string of length n"""

    return "".join(random.choice("0123456789abcdef") for i in range(length))


def random_digits(length: int) -> str:
    """Returns a random string of digits of length n"""

    return "".join(random.choice("0123456789") for i in range(length))


def get_nyanko_auth_url() -> str:
    """Returns the url for the nyanko auth server"""

    return "https://nyanko-auth.ponosgames.com"


def generate_nyanko_signature(inquiry_code: str, data: str) -> str:
    """Generates a signature for the given data with the given inquiry_code"""

    inquiry_code = inquiry_code.encode("utf-8")
    data = data.encode("utf-8")
    random_data = random_hex(64).encode("utf-8")
    input_rand = inquiry_code + random_data
    hmac_data = hmac.new(input_rand, data, hashlib.sha256).hexdigest()
    final_signature = random_data.decode("utf-8") + hmac_data
    return final_signature


def confirm_nyanko_signature(signature: str, json_data: str, inquiry_code: str) -> bool:
    """Checks to see if a given signature is valid"""

    random_data = signature[:64]
    hmac_data = signature[64:]
    input_rand = inquiry_code.encode("utf-8") + random_data.encode("utf-8")
    new_hmac = hmac.new(
        input_rand, json_data.encode("utf-8"), hashlib.sha256
    ).hexdigest()
    return new_hmac == hmac_data


def get_headers(inquiry_code: str, data: str) -> dict:
    """Returns the headers for the given inquiry_code and data"""

    headers = {
        "accept-encoding": "gzip",
        "connection": "keep-alive",
        "content-type": "application/json",
        "nyanko-signature": generate_nyanko_signature(inquiry_code, data),
        "nyanko-signature-algorithm": "HMACSHA256",
        "nyanko-signature-version": "1",
        "nyanko-timestamp": str(get_current_time_stamp()),
    }
    return headers


def get_password(inquiry_code: str) -> Union[dict, None]:
    """Returns the password for the given inquiry_code"""

    url = get_nyanko_auth_url() + "/v1/users"
    data = {
        "accountCode": inquiry_code,
        "accountCreatedAt": str(get_current_time_stamp() - 5),
        "nonce": random_hex(32),
    }
    data = json.dumps(data)
    data = data.replace(" ", "")
    headers = get_headers(inquiry_code, data)
    response = requests.post(url, data=data, headers=headers)
    data = response.json()
    if data["statusCode"] != 1:
        return None
    payload = data["payload"]
    return payload


def get_password_refresh(
    inquiry_code: str, password_refresh_token: str
) -> Union[dict, None]:
    """Returns the password for the given inquiry_code"""

    url = get_nyanko_auth_url() + "/v1/user/password"
    data = {
        "accountCode": inquiry_code,
        "nonce": random_hex(32),
        "passwordRefreshToken": password_refresh_token,
    }
    data = json.dumps(data)
    data = data.replace(" ", "")
    headers = get_headers(inquiry_code, data)
    try:
        response = requests.post(url, data=data, headers=headers)
    except requests.exceptions.RequestException:
        helper.colored_text(
            "Error couldn't get account password token.\nCheck your internet connection",
            helper.RED,
        )
        return False
    data = response.json()
    if data["statusCode"] != 1:
        return None
    payload = data["payload"]
    return payload


def get_client_info(country_code: str, game_version: str) -> dict:
    """Returns the client info for the given country_code and game_version"""

    country_code = country_code.replace("jp", "ja")

    data = {
        "clientInfo": {
            "client": {"countryCode": country_code, "version": game_version},
            "device": {"model": "SM-G973N"},
            "os": {"type": "android", "version": "7.1.2"},
        },
        "nonce": random_hex(32),
    }
    return data


def get_token(inquiry_code: str, password: str, save_stats: dict) -> Union[None, str]:
    """Returns the token for the given inquiry_code and password"""

    url = get_nyanko_auth_url() + "/v1/tokens"
    data = get_client_info(save_stats["version"], save_stats["game_version"])
    data["password"] = password
    data["accountCode"] = inquiry_code
    data = json.dumps(data)
    data = data.replace(" ", "")
    headers = get_headers(inquiry_code, data)
    response = requests.post(url, data=data, headers=headers)
    data = response.json()
    if data["statusCode"] != 1:
        return None
    payload = data["payload"]
    return payload["token"]


def download_save(
    country_code: str, transfer_code: str, confirmation_code: str, game_version: str
) -> Union[None, bytes]:
    """Downloads the save for the given country_code, transfer_code, confirmation_code and game_version"""

    if country_code == "jp":
        country_code = "ja"
    url = f"https://nyanko-save.ponosgames.com/v1/transfers/{transfer_code}/reception"
    data = get_client_info(country_code, game_version)
    data["pin"] = confirmation_code
    data = json.dumps(data)
    data = data.replace(" ", "")
    headers = headers = {"content-type": "application/json", "accept-encoding": "gzip"}
    try:
        response = requests.post(url, data=data, headers=headers)
    except requests.exceptions.RequestException:
        helper.colored_text(
            "Unable to download save data\nCheck your internet connection",
            base=helper.RED,
        )
        return
    return response.content


def download_save_handler() -> Union[None, str]:
    """Handler for the download_save function"""

    country_code = input("Enter your game version (en, ja, kr, tw):")
    if country_code == "jp":
        country_code = "ja"
    transfer_code = input("Enter transfer code:")
    confirmation_code = input("Enter confirmation code:")
    game_version = input("Enter current game version (e.g 11.3, 9.6, 10.4.0):")
    game_version = helper.str_to_gv(game_version)
    save_data = download_save(
        country_code, transfer_code, confirmation_code, game_version
    )
    if not save_data:
        return None
    try:
        save_data = json.loads(save_data)
    except (json.decoder.JSONDecodeError, UnicodeDecodeError):
        helper.colored_text("Successfully downloaded save data\n", base=helper.GREEN)
    else:
        helper.colored_text(
            "Incorrect transfer code / confirmation code / game version",
            base=helper.RED,
        )
        return None
    path = helper.save_file(
        "Save save data", helper.get_save_file_filetype(), "SAVE_DATA"
    )
    helper.write_file_bytes(path, save_data)
    return path


def upload_save_data(
    token: str, inquiry_code: str, save_data: bytes, save_stats: dict
) -> Union[None, dict]:
    """Uploads the save data for the given token, inquiry_code and save_data"""

    managed_item_details = {
        "managedItemDetails": [],
        "nonce": random_hex(32),
        "playTime": helper.time_to_frames(save_stats["play_time"]),
        "rank": 11,
        "receiptLogIds": [],
    }
    managed_item_details = json.dumps(managed_item_details)
    managed_item_details = managed_item_details.replace(" ", "")
    headers = get_headers(inquiry_code, managed_item_details)
    headers["authorization"] = "Bearer " + token
    boundary = f"__-----------------------{random_digits(9)}-2147483648".encode("utf-8")
    headers["content-type"] = "multipart/form-data; boundary=" + boundary.decode(
        "utf-8"
    )

    body = b"--" + boundary + b"\r\n"
    body += b'Content-Disposition: form-data; name="saveData"; filename="data.sav"\r\n'
    body += b"Content-Type: application/octet-stream\r\n\r\n"
    body += save_data + b"\r\n"
    body += b"--" + boundary + b"\r\n"
    body += b'Content-Disposition: form-data; name="metaData"\r\n\r\n'
    body += managed_item_details.encode("utf-8") + b"\r\n"
    body += b"--" + boundary + b"--\r\n"

    url = "https://nyanko-save.ponosgames.com/v1/transfers"
    response = requests.post(url, data=body, headers=headers)
    data = response.json()
    if data["statusCode"] != 1:
        print("Error: " + response.text)
        return None
    return data["payload"]


def get_inquiry_code() -> str:
    """Returns a new inquiry code"""

    url = "https://nyanko-backups.ponosgames.com/?action=createAccount&referenceId="
    response = requests.get(url)
    data = response.json()
    return data["accountId"]


def check_password(
    save_stats: dict, get_pass: bool = False
) -> Union[dict, tuple[dict, dict]]:
    """Checks the password for the given save_stats and generates new if stuff is invalid"""

    inquiry_code = save_stats["inquiry_code"]
    password_refresh_token = save_stats["token"]

    helper.colored_text("Getting your account password token...", helper.GREEN)
    password_refresh_data = get_password_refresh(inquiry_code, password_refresh_token)

    if password_refresh_data is False:
        return save_stats

    if password_refresh_data is None:
        helper.colored_text(
            "Error getting your account password token, Generating a new one...",
            helper.RED,
        )
        password_refresh_data = get_password(inquiry_code)

        if password_refresh_data is None:
            helper.colored_text(
                "Your inquiry code seems to be invalid. Generating a new one...",
                helper.RED,
            )
            inquiry_code = get_inquiry_code()

            save_stats["inquiry_code"] = inquiry_code
            helper.colored_text(
                f"A new account code has been generated for your account: &{inquiry_code}&",
                helper.GREEN,
                new=helper.WHITE,
            )

            return check_password(save_stats, get_pass)
        if "accountCode" in password_refresh_data:
            inquiry_code = password_refresh_data["accountCode"]
            save_stats["inquiry_code"] = inquiry_code

            helper.colored_text(
                f"A new account code has been generated for your account: &{inquiry_code}&",
                helper.GREEN,
                new=helper.WHITE,
            )

        password_refresh_token = password_refresh_data["passwordRefreshToken"]
        save_stats["token"] = password_refresh_token
        helper.colored_text(
            f"A new account password token has been generated for your account : &{password_refresh_token}&",
            helper.GREEN,
            helper.WHITE,
        )
    if get_pass:
        return save_stats, password_refresh_data
    return save_stats


def upload_handler(save_stats: dict, path: str) -> dict:
    """Handles the upload of the save data"""

    save_stats, password_refresh_data = check_password(save_stats, True)
    inquiry_code = save_stats["inquiry_code"]

    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    save_data = helper.write_file_bytes(path, save_data)

    helper.colored_text("Getting account auth token...", helper.GREEN)
    token = get_token(inquiry_code, password_refresh_data["password"], save_stats)

    helper.colored_text("Uploading save data...", helper.GREEN)
    upload_data = upload_save_data(token, inquiry_code, save_data, save_stats)

    helper.colored_text(
        "After entering these codes in game, you may get a ban message when pressing play. If you do, just press play again and it should go away.\nPress enter to get your codes...",
        helper.DARK_YELLOW,
    )
    input()
    return upload_data



"""Helper script for usefull functions"""

import json
import os
import sys
from tkinter import Tk, filedialog
from typing import Union

import colored

from . import parse_save, patcher, serialise_save, user_input_handler

root = Tk()
root.withdraw()

GREEN = "#008000"
RED = "#FF0000"
DARK_YELLOW = "#D7C32A"
BLACK = "#000000"
WHITE = "#FFFFFF"
CYAN = "#00FFFF"


def str_to_gv(game_version: str) -> str:
    """Turn a game version with semantic versioning to integer representation"""

    split_gv = game_version.split(".")
    if len(split_gv) == 2:
        split_gv.append("0")
    final = ""
    for split in split_gv:
        final += split.zfill(2)
    return final


def gv_to_str(game_version: int) -> str:
    """Turn a game version integer into a string"""

    return f"{game_version[0:2]}.{game_version[2:4]}.{game_version[4:6]}"


def check_clamp(
    values: list, max_value: int, min_value: int = 0, offset: int = -1
) -> list:
    """# turn a list of strings into a list of ints and clamp them between a min and max"""

    if isinstance(values, str):
        values = [values]
    int_values = []
    for value in values:
        value = str(value).strip(" ")
        value = check_int(value)
        if value is None:
            continue
        value = clamp(value, min_value, max_value)
        value += offset
        int_values.append(value)
    return int_values


def num_to_bytes(num: int, length: int) -> bytes:
    """Turn number into little endian bytes"""

    return num.to_bytes(length, byteorder="little")


def load_json_handler(json_path: str) -> Union[None, str]:
    """Load a save_data json file and serialise it"""

    save_stats = load_json(json_path)
    save_data = serialise_save.start_serialize(save_stats)
    save_data = patcher.patch_save_data(save_data, save_stats["version"])
    path = save_file(
        "Save file",
        get_save_file_filetype(),
        os.path.join(os.path.dirname(json_path), "SAVE_DATA"),
    )
    if not path:
        return None
    write_file_bytes(path, save_data)
    return path


def load_json(json_path: str) -> dict:
    """Load a json file"""

    return json.loads(read_file_string(json_path))

def check_cat_ids(cat_ids: list, save_stats: dict) -> list:
    """Check if a list of cat ids is valid"""

    for cat_id in cat_ids:
        if cat_id > len(save_stats["cats"]):
            print(
                colored_text(
                    f"Invalid cat id {cat_id}", base=RED
                )
            )
            cat_ids.remove(cat_id)
    return cat_ids


def check_int(value: str) -> Union[int, None]:
    """Check if a string is an integer"""

    value = str(value).strip(" ")
    try:
        return int(value)
    except ValueError:
        return None


def set_range(
    list_original: Union[bytes, list], values: list, start_point: int
) -> bytes:
    """Set a range of values in a list"""

    list_original = list(list_original)
    list_original[start_point : start_point + len(values)] = values
    return bytes(list_original)


def frames_to_time(frames: int) -> dict:
    """Turn frames into hours, minutes, seconds, frames"""

    seconds, frames = divmod(frames, 30)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds, "frames": frames}


def seconds_to_time(seconds: int) -> dict:
    """Turn seconds into hours, minutes, seconds"""

    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return {"hh": hours, "mm": minutes, "ss": seconds}


def time_to_seconds(time: dict) -> int:
    """Turn hours, minutes, seconds into seconds"""
    seconds = time["ss"]
    seconds += time["mm"] * 60
    seconds += time["hh"] * 60 * 60
    return seconds


def time_to_frames(time: dict) -> int:
    """Turn hours, minutes, seconds, frames into frames"""

    total_frames = time["frames"]
    total_frames += time["ss"] * 30
    total_frames += time["mm"] * 60 * 30
    total_frames += time["hh"] * 60 * 60 * 30
    return total_frames


def chunks(lst: list, chunk_len: int) -> list:
    """Split list into chunks of n"""

    for i in range(0, len(lst), chunk_len):
        yield lst[i : i + chunk_len]


def read_file_bytes(file_path: str) -> bytes:
    """Read file as bytes"""

    with open(file_path, "rb") as file:
        return file.read()


def write_file_bytes(file_path: str, data: bytes) -> bytes:
    """Write file as bytes"""

    try:
        with open(file_path, "wb") as file:
            file.write(data)
    except PermissionError:
        print(colored_text("Permission denied. Make sure the file is not in use", base=RED))
        sys.exit(1)
    return data


def write_file_string(file_path: str, data: str):
    """Write file as string"""

    with open(file_path, "w", encoding="utf-8") as file:
        file.write(data)


def clamp(value: int, min_value: int, max_value: int) -> int:
    """Clamp a value between two values"""

    return max(min(value, max_value), min_value)


def get_country_code(save_data: dict) -> str:
    """Ask the user for their country code if it cannot be detected"""

    country_code = patcher.detect_game_version(save_data)
    if not country_code:
        country_code = user_input_handler.colored_input(
            "Enter your game version (&en&, &jp&, &kr&, &tw&):"
        )
    return country_code


def load_save_file(path: str) -> dict:
    """Load a save file, get the country code, create a backup and parse the save data"""

    save_data = read_file_bytes(path)
    country_code = get_country_code(save_data)
    colored_text(f"Game version: &{country_code}&")
    save_stats = parse_save.start_parse(save_data, country_code)
    write_file_bytes(path + "_backup", save_data)
    colored_text(f"Backup created at: &{os.path.abspath(path + '_backup')}&", new=GREEN)
    return {
        "save_data": save_data,
        "country_code": country_code,
        "save_stats": save_stats,
    }


def offset_list(lst: list, offset: int) -> list:
    """Offset each value in an list by a certain amount"""

    new_list = []
    for item in lst:
        new_list.append(item + offset)
    return new_list


def copy_first_n(lst: list, number: int) -> list:
    """Get the nth item in a list of lists"""

    new_list = []
    for item in lst:
        new_list.append(item[number])
    return new_list


def write_save_data(save_data: dict, country_code: str, path: str, prompt: bool):
    """Patch the save data and write it"""

    save_data = patcher.patch_save_data(save_data, country_code)
    if prompt:
        path = save_file("Save File", get_save_file_filetype(), path)
        if not path:
            colored_text("Save cancelled", new=RED)
            return
    write_file_bytes(path, save_data)
    colored_text(f"Saved to: &{os.path.abspath(path)}&", new=GREEN)
    return save_data


def save_file(title: str, file_types: list[tuple[str]], path: str) -> str:
    """Save a file with tkinter"""

    try:
        path = filedialog.asksaveasfile(
            "w",
            confirmoverwrite=True,
            initialfile=os.path.basename(path),
            filetypes=file_types,
            title=title,
        )
    except PermissionError:
        print(colored_text("Permission denied. Make sure the file is not in use", base=RED))
        sys.exit(1)
    if not path:
        return ""
    return path.name


def select_file(title: str, file_types: list[tuple[str]], default_dir: str = "") -> str:
    """Select a file with tkinter"""
    file_path = filedialog.askopenfilename(
        initialdir=default_dir, title=title, filetypes=file_types
    )
    return file_path


def get_save_file_filetype() -> list[tuple[str]]:
    """Get the file types for the save file"""

    return [("Battle Cats Save Files", "*SAVE_DATA*"), ("All Files", "*.*")]


def colored_list(
    items: list, extra_data: list = None, index: bool = True, offset: int = None
):
    """Print a list with colors and extra data if provided"""

    final = ""
    for i, item in enumerate(items):
        if index:
            final += f"{i+1}. "
        final += f"&{item}&"
        if extra_data:
            if extra_data[i] is not None:
                if isinstance(offset, int):
                    final += f" &:& {extra_data[i]+offset}"
                else:
                    final += f" &:& {extra_data[i]}"
        final += "\n"
    final = final.rstrip("\n")
    colored_text(final)


def int_to_str_ls(int_list: list) -> list:
    """Turn list of ints to list of strings"""

    str_list = []
    for i in int_list:
        str_list.append(str(i))
    return str_list


def get_file(file_name: str) -> str:
    """Get file in files folder"""

    file_path = os.path.join(get_local_files_path(), file_name)
    return file_path


def read_file_string(file_path: str) -> str:
    """Read file as string"""
    try:
        with open(file_path, "r", encoding="utf-8") as file:
            return file.read()
    except UnicodeDecodeError:
        colored_text("Error: make sure the file is in UTF-8 encoding", base=RED)
        sys.exit(1)

def get_local_files_path() -> str:
    """Get the local files path"""

    dir_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "files")
    return dir_path


def colored_text(
    text: str,
    base: str = WHITE,
    new: str = DARK_YELLOW,
    split_char: str = "&",
    end: str = "\n",
):
    """Print text with colors"""
    color_new = colored.fg(new)
    color_base = colored.fg(base)
    color_reset = colored.fg(WHITE)

    text_split = text.split(split_char)
    for i, text_section in enumerate(text_split):
        if i % 2:
            print(f"{color_new}{text_section}{color_base}", end="")
        else:
            print(f"{color_base}{text_section}{color_base}", end="")
    print(color_reset, end=end)


def remove_empty_items(list_of_lists: list[list]) -> list[list]:
    """Remove empty items from list of lists"""

    for i, lst in enumerate(list_of_lists):
        list_of_lists[i] = list(filter(None, lst))
    list_of_lists = remove_empty_lists(list_of_lists)
    return list_of_lists


def remove_empty_lists(list_of_lists: list[list]) -> list[list]:
    """Remove empty lists from list of lists"""

    new_lists = []
    for lst in list_of_lists:
        if lst:
            new_lists.append(lst)
    return new_lists


def parse_int_list_list(list_of_lists: list[list[str]]) -> list[list[int]]:
    """Turn list of list of strings into list of list of ints"""

    for i, lst in enumerate(list_of_lists):
        for j, item in enumerate(lst):
            try:
                list_of_lists[i][j] = int(item)
            except ValueError:
                continue
    return list_of_lists


def parse_int_list(lst: list, offset: int) -> list[int]:
    """Turn string list to int list"""

    new_list = []
    for item in lst:
        try:
            new_list.append(int(item) + offset)
        except ValueError:
            pass
    return new_list



"""Checks if the editor is updated and if not, updates it"""

import subprocess
import sys
from typing import Union
import requests
from . import helper, user_input_handler

PKG_NAME = "battle-cats-save-editor"


def get_local_version() -> str:
    """Get the local version of the program"""

    return helper.read_file_string(helper.get_file("version.txt"))


def get_pypi_version() -> Union[None, str]:
    """Get latest pypi version of the program"""

    try:
        response = requests.get(f"https://pypi.python.org/pypi/{PKG_NAME}/json")
        response.raise_for_status()
        return response.json()["info"]["version"]
    except requests.exceptions.RequestException:
        helper.colored_text(
            "Could not get latest version from pypi\nCheck your intnet connection\n",
            base=helper.RED,
        )
        return None


def update(command: str = "py"):
    """Update pypi package testing for py and python"""

    helper.colored_text("Updating...", base=helper.GREEN)
    try:
        subprocess.run(
            f"{command} -m pip install --upgrade {PKG_NAME}",
            shell=True,
            capture_output=True,
            check=True,
        )
        helper.colored_text("Update successful", base=helper.GREEN)
    except subprocess.CalledProcessError as err:
        helper.colored_text("Update failed", base=helper.RED)
        if command == "py":
            helper.colored_text("Trying with python instead", base=helper.RED)
            update("python")
        else:
            helper.colored_text(
                f"Error: {err.stderr.decode('utf-8')}\nYou may need to manually update with py -m pip install -U {PKG_NAME}",
                base=helper.RED,
            )


def check_update():
    """Check for update and allow user to automatically update"""

    local_version = get_local_version()
    pypi_version = get_pypi_version()
    if pypi_version is not None and pypi_version > local_version:
        helper.colored_text(
            f"Update available: &{local_version}& -> &{pypi_version}&",
            base=helper.GREEN,
            new=helper.WHITE,
        )
        usr_update = user_input_handler.colored_input("Update now? (&y&/&n&):")
        if usr_update.lower()[0] == "y":
            update()
            helper.colored_text(
                "Please rerun the program to use the latest update", base=helper.GREEN
            )
            sys.exit()
    else:
        helper.colored_text(
            f"No update available (local: &{local_version}& pypi: &{pypi_version}&)",
            base=helper.GREEN,
            new=helper.WHITE,
        )



"""Fix the elsewhere issue and unban an account"""

import json
import os

from ... import adb_handler, helper, server_handler


def edit_cache(password: str, token: str, save_stats: str):
    """Edit the cache file in /data/data/jp.co.ponos.battlecats/files/cache/ to add the token and password"""

    data = {"password": password, "token": token}
    data = json.dumps(data)
    data = data.replace(" ", "")
    inquiry_code = save_stats["inquiry_code"]
    local_path = os.path.abspath(inquiry_code + ".json")

    helper.write_file_string(local_path, data)
    game_v = save_stats["version"]
    if game_v == "jp":
        game_v = ""
    success = (
        adb_handler.adb_cmd_handler(
            f'adb shell mv "{local_path}" "/data/data/jp.co.ponos.battlecats{game_v}/cache/{inquiry_code}.json"'
        )
        is not None
    )
    os.remove(local_path)
    return success


def fix_elsewhere(save_stats: dict) -> dict:
    """Handler for fixing the elsewhere issue and unban an account"""

    save_stats, password_refresh_data = server_handler.check_password(save_stats, True)
    inquiry_code = save_stats["inquiry_code"]
    helper.colored_text("Getting account auth token...", helper.GREEN)
    token = server_handler.get_token(
        inquiry_code, password_refresh_data["password"], save_stats
    )
    if edit_cache(password_refresh_data["password"], token, save_stats):
        helper.colored_text(
            "Done!\nYou may get a ban message when pressing play. If you do, just press play again and it should go away\nPress enter to continue...",
            helper.DARK_YELLOW,
        )
        input()
    return save_stats



"""Handler for basic, items that work in a common way"""
from ... import helper, item


def edit_cat_food(save_stats: dict) -> dict:
    """Handler for editing cat food"""

    cat_food = item.Item(
        name="Cat Food",
        value=save_stats["cat_food"]["Value"],
        max_value=45000,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=False,
        ),
    )
    cat_food.edit()
    save_stats["cat_food"]["Value"] = cat_food.value
    return save_stats


def edit_xp(save_stats: dict) -> dict:
    """Handler for editing xp"""

    experience = item.Item(
        name="XP",
        value=save_stats["xp"]["Value"],
        max_value=99999999,
        edit_name="value",
    )
    experience.edit()
    save_stats["xp"]["Value"] = experience.value
    return save_stats


def edit_normal_tickets(save_stats: dict) -> dict:
    """Handler for editing normal tickets"""

    normal_tickets = item.Item(
        name="Normal Tickets",
        value=save_stats["normal_tickets"]["Value"],
        max_value=2999,
        edit_name="value",
    )
    normal_tickets.edit()
    save_stats["normal_tickets"]["Value"] = normal_tickets.value
    return save_stats


def edit_rare_tickets(save_stats: dict) -> dict:
    """Handler for editing rare tickets"""

    rare_tickets = item.Item(
        name="Rare Tickets",
        value=save_stats["rare_tickets"]["Value"],
        max_value=299,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=True,
            work_around_text='&Instead of editing rare tickets directly, use the "Normal Ticket Max Trade Progress" conversion feature instead! It is much more safe.',
        ),
    )
    rare_tickets.edit()
    save_stats["rare_tickets"]["Value"] = rare_tickets.value
    return save_stats


def edit_platinum_tickets(save_stats: dict) -> dict:
    """Handler for editing platinum tickets"""

    platinum_tickets = item.Item(
        name="Platinum Tickets",
        value=save_stats["platinum_tickets"]["Value"],
        max_value=9,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=True,
            work_around_text="&Instead of editing platinum tickets, edit platinum shards instead! They are much more safe. 10 platinum shards = 1 platinum ticket",
        ),
    )
    platinum_tickets.edit()
    save_stats["platinum_tickets"]["Value"] = platinum_tickets.value
    return save_stats


def edit_platinum_shards(save_stats: dict) -> dict:
    """Handler for editing platinum shards"""

    platinum_shards = item.Item(
        name="Platinum Shards",
        value=save_stats["platinum_shards"]["Value"],
        max_value=999,
        edit_name="value",
    )
    platinum_shards.edit()
    save_stats["platinum_shards"]["Value"] = platinum_shards.value
    return save_stats


def edit_np(save_stats: dict) -> dict:
    """Handler for editing np"""

    nyanko_points = item.Item(
        name="NP",
        value=save_stats["np"]["Value"],
        max_value=9999,
        edit_name="value",
    )
    nyanko_points.edit()
    save_stats["np"]["Value"] = nyanko_points.value
    return save_stats


def edit_leadership(save_stats: dict) -> dict:
    """Handler for editing leadership"""

    leadership = item.Item(
        name="Leadership",
        value=save_stats["leadership"]["Value"],
        max_value=9999,
        edit_name="value",
    )
    leadership.edit()
    save_stats["leadership"]["Value"] = leadership.value
    return save_stats


def edit_battle_items(save_stats: dict) -> dict:
    """Handler for editing battle items"""

    battle_items = item.create_item_group(
        names=[
            "Speed Up",
            "Treasure Radar",
            "Rich Cat",
            "Cat CPU",
            "Cat Jobs",
            "Sniper the Cat",
        ],
        values=save_stats["battle_items"],
        maxes=9999,
        edit_name="value",
        group_name="Battle Items",
    )
    battle_items.edit()
    save_stats["battle_items"] = battle_items.values

    return save_stats


def edit_catseyes(save_stats: dict) -> dict:
    """Handler for editing catseyes"""

    catseyes = item.create_item_group(
        names=[
            "Special",
            "Rare",
            "Super Rare",
            "Uber Super Rare",
            "Legend Rare",
        ],
        values=save_stats["catseyes"],
        maxes=9999,
        edit_name="value",
        group_name="Catseyes",
    )
    catseyes.edit()
    save_stats["catseyes"] = catseyes.values
    return save_stats


def edit_catfruit(save_stats: dict) -> dict:
    """Handler for editing catruit"""

    max_cf = 128
    if save_stats["game_version"]["Value"] >= 110400:
        max_cf = None

    catfruit = item.create_item_group(
        names=[
            "Purple Seeds",
            "Red Seeds",
            "Blue Seeds",
            "Green Seeds",
            "Yellow Seeds",
            "Purple Fruit",
            "Red Fruit",
            "Blue Fruit",
            "Green Fruit",
            "Yellow Fruit",
            "Epic Fruit",
            "Elder Seeds",
            "Elder Fruit",
            "Epic Seeds",
            "Gold Fruit",
            "Aku Seeds",
            "Aku Fruit",
            "Gold Seeds",
            "Purple Behemoth Stone",
            "Red Behemoth Stone",
            "Blue Behemoth Stone",
            "Green Behemoth Stone",
            "Yellow Behemoth Stone",
        ],
        values=save_stats["cat_fruit"],
        maxes=max_cf,
        edit_name="value",
        group_name="Catfruit",
    )
    catfruit.edit()
    save_stats["cat_fruit"] = catfruit.values
    return save_stats


def edit_engineers(save_stats: dict) -> dict:
    """Handler for editing ototo engineers"""

    engineers = item.Item(
        name="Ototo Engineers",
        value=save_stats["engineers"]["Value"],
        max_value=4,
        edit_name="value",
    )
    engineers.edit()
    save_stats["engineers"]["Value"] = engineers.value
    return save_stats


def edit_base_materials(save_stats: dict) -> dict:
    """Handler for editing ototo base materials"""

    base_materials = item.create_item_group(
        names=[
            "Bricks",
            "Feathers",
            "Coal",
            "Sprockets",
            "Gold",
            "Meteorite",
            "Beast Bones",
            "Ammonite",
        ],
        values=save_stats["base_materials"],
        maxes=9999,
        edit_name="value",
        group_name="Base Materials",
    )
    base_materials.edit()
    save_stats["base_materials"] = base_materials.values
    return save_stats


def edit_catamins(save_stats: dict) -> dict:
    """Handler for editing catamins"""

    catamins = item.create_item_group(
        names=[
            "Catamin A",
            "Catamin B",
            "Catamin C",
        ],
        values=save_stats["catamins"],
        maxes=9999,
        edit_name="value",
        group_name="Catamins",
    )
    catamins.edit()
    save_stats["catamins"] = catamins.values
    return save_stats


def edit_inquiry_code(save_stats: dict) -> dict:
    """Handler for editing the inquiry code"""

    print(
        "WARNING: Editing your inquiry code should only be done if you know what you are doing! Because it will lead to an elsewhere error in-game if not done correctly!"
    )
    inquiry_code = item.Item(
        name="Inquiry Code",
        value=save_stats["inquiry_code"],
        max_value=None,
        edit_name="value",
    )
    inquiry_code.edit()
    save_stats["inquiry_code"] = inquiry_code.value
    return save_stats


def edit_rare_gacha_seed(save_stats: dict) -> dict:
    """Handler for editing the rare gacha seed"""

    rare_gacha_seed = item.Item(
        name="Rare Gacha Seed",
        value=save_stats["rare_gacha_seed"]["Value"],
        max_value=None,
        edit_name="value",
    )
    rare_gacha_seed.edit()
    save_stats["rare_gacha_seed"]["Value"] = rare_gacha_seed.value
    return save_stats


def edit_unlocked_slots(save_stats: dict) -> dict:
    """Handler for editing the amount of unlocked slots"""

    unlocked_slots = item.Item(
        name="Unlocked Slots",
        value=save_stats["unlocked_slots"]["Value"],
        max_value=len(save_stats["slots"]),
        edit_name="value",
    )
    unlocked_slots.edit()
    save_stats["unlocked_slots"]["Value"] = unlocked_slots.value
    return save_stats


def edit_token(save_stats: dict) -> dict:
    """Handler for editing the password-refresh-token"""

    print(
        "WARNING: Editing your token should only be done if you know what you are doing! Because it will lead to an elsewhere error in-game if not done correctly!"
    )
    token = item.Item(
        name="Token",
        value=save_stats["token"],
        max_value=None,
        edit_name="value",
    )
    token.edit()
    save_stats["token"] = token.value
    return save_stats


def edit_restart_pack(save_stats: dict) -> dict:
    """Handler for giving the restart pack"""

    save_stats["restart_pack"]["Value"] = 1
    print("Successfully gave the restart pack")
    return save_stats


def edit_challenge_battle(save_stats: dict) -> dict:
    """Handler for editing the score of the challenge battle"""

    challenge_battle = item.Item(
        name="Challenge Battle",
        value=save_stats["challenge"]["Score"]["Value"],
        max_value=None,
        edit_name="score",
    )
    challenge_battle.edit()
    save_stats["challenge"]["Score"]["Value"] = challenge_battle.value
    save_stats["challenge"]["Cleared"]["Value"] = 1
    return save_stats


def edit_legend_tickets(save_stats: dict) -> dict:
    """Handler for editing legend tickets"""

    legend_tickets = item.Item(
        name="Legend Tickets",
        value=save_stats["legend_tickets"]["Value"],
        max_value=4,
        edit_name="value",
        bannable=item.BannableItem(
            is_bannable=True,
            has_workaround=False,
        ),
    )
    legend_tickets.edit()
    save_stats["legend_tickets"]["Value"] = legend_tickets.value
    return save_stats


def edit_dojo_score(save_stats: dict) -> dict:
    """Handler for editing the dojo score"""

    if not save_stats["dojo_data"]:
        helper.colored_text(
            "No catclaw dojo data found\nPlease enter the catclaw dojo menu and try again",
            base=helper.RED,
        )
        return save_stats

    dojo_score = item.Item(
        name="Dojo Score",
        value=save_stats["dojo_data"][0][0],
        max_value=None,
        edit_name="value",
    )
    dojo_score.edit()
    save_stats["dojo_data"][0][0] = dojo_score.value
    return save_stats



"""Handler for clearing the tutorial"""


def clear_tutorial(save_stats: dict) -> dict:
    """Handler for clearing the tutorial"""

    save_stats["tutorial_cleared"]["Value"] = 1
    if save_stats["story_chapters"]["Chapter Progress"][0] == 0:
        save_stats["story_chapters"]["Chapter Progress"][0] = 1
    save_stats["story_chapters"]["Times Cleared"][0][0] = 1
    print("Successfully cleared the tutorial")
    return save_stats



"""Handler for editing main story treasures"""

from ..levels import main_story
from ... import helper, csv_file_handler, user_input_handler, item


def get_stages() -> list:
    """Get what stages belong to which treasure group"""

    treasures_values = []
    eoc_treasures = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/treasureData0.csv")
    )[11:22]
    itf_treasures = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/treasureData1.csv")
    )[11:22]
    cotc_treasures = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/treasureData2_0.csv")
    )[11:22]

    treasures_values.append(remove_negative_1(eoc_treasures))
    treasures_values.append(remove_negative_1(itf_treasures))
    treasures_values.append(remove_negative_1(cotc_treasures))
    return treasures_values


def get_treasure_groups() -> dict:
    """Get the names and stages of all of the treasure groups"""

    treasure_stages = get_stages()
    treasure_names = get_names()
    return {"names": treasure_names, "stages": treasure_stages}


def get_names() -> list:
    """Get the names of all of the treasure groups"""

    names = []
    eoc_names = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/Treasure3_0_en.csv"), delimiter="|"
    )[:11]
    itf_names = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/Treasure3_1_AfterFirstEncounter_en.csv"),
        delimiter="|",
    )[:11]
    cotc_names = csv_file_handler.parse_csv(
        helper.get_file("game_data/treasures/Treasure3_2_0_en.csv"), delimiter="|"
    )[:11]

    names.append(helper.copy_first_n(eoc_names, 0))
    names.append(helper.copy_first_n(itf_names, 0))
    names.append(helper.copy_first_n(cotc_names, 0))

    return names


def remove_negative_1(data: list) -> list:
    """Remove items from a list that have a negative value of 1"""

    new_data = data.copy()
    for i, val in enumerate(data):
        if -1 in val:
            new_data[i] = new_data[i][:-1]
    return new_data


def treasure_groups(save_stats: dict) -> dict:
    """Handler for editing treasure groups"""

    treasure_grps = get_treasure_groups()
    treasures_stats = save_stats["treasures"]

    helper.colored_list(main_story.chapters)
    ids = user_input_handler.colored_input(
        "Enter a number from 1 to 9 (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")

    ids = helper.check_clamp(ids, 9, 1, -1)

    for chapter_id in ids:
        print(f"Chapter: {main_story.chapters[chapter_id]}")
        type_id = chapter_id // 3
        if chapter_id > 2:
            chapter_id += 1
        names = treasure_grps["names"][type_id]
        treasure_levels = [-1] * len(names)
        print("0 = None, 1 = Inferior, 2 = Normal, 3 = Superior")
        treasure_levels = item.create_item_group(
            names=names,
            values=None,
            maxes=None,
            edit_name="treasure level",
            group_name="Treasures",
        )
        treasure_levels.edit()

        for i, treasure_level in enumerate(treasure_levels.values):
            if treasure_level is None:
                continue
            stages = treasure_grps["stages"][type_id][i]
            for stage in stages:
                treasures_stats[chapter_id][stage] = treasure_level
    save_stats["treasures"] = treasures_stats

    return save_stats


def specific_treasures(save_stats: dict) -> dict:
    """Handler for editing treasure levels"""

    individual = (
        user_input_handler.colored_input(
            "Do you want to edit the treasures for individual levels &(1)&, or groups of treasures (e.g energy drink, aqua crystal) &(2)&?:"
        )
        == "1"
    )
    if not individual:
        return treasure_groups(save_stats)
    treasure_stats = save_stats["treasures"]

    helper.colored_list(main_story.chapters)
    ids = user_input_handler.colored_input(
        "Enter a number from 1 to 9 (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")

    ids = user_input_handler.create_all_list(ids, 9)
    ids = helper.check_clamp(ids, 9, 1, -1)

    for chapter_id in ids:
        stage_ids = user_input_handler.get_range(
            user_input_handler.colored_input(
                f"Enter stage ids for chapter {main_story.chapters[chapter_id]}(e.g 1=korea, 2=mongolia)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
            ),
            48,
        )
        stage_ids = helper.check_clamp(stage_ids, 48, 1, -1)
        all_at_once = len(stage_ids) == 48
        treasure_data = [-1] * 49
        treasure_data = user_input_handler.handle_all_at_once(
            stage_ids,
            all_at_once,
            treasure_data,
            range(1, 50),
            "treasure level",
            "stage",
            "(&0&=none, &1&=inferior, &2&=normal, &3&=superior)",
        )
        if chapter_id > 2:
            chapter_id += 1
        for i, stage in enumerate(treasure_data):
            if stage == -1:
                continue
            if i > 45:
                stage_id = i
            else:
                stage_id = 45 - i
            treasure_stats[chapter_id][stage_id] = stage

    save_stats["treasures"] = treasure_stats
    print("Successfully set treasures")
    return save_stats


def treasures(save_stats: dict) -> dict:
    """Handler for editing treasure chapters"""

    treasure_stats = save_stats["treasures"]

    helper.colored_list(main_story.chapters)
    ids = user_input_handler.colored_input(
        "10. &All at once&\nEnter a number from 1 to 10 (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")

    data = user_input_handler.create_all_list(ids, 10, True)
    ids = data["ids"]
    ids = helper.check_clamp(ids, 10, 1, -1)
    all_at_once = data["at_once"]
    usr_levels = [-1] * 9
    usr_levels = user_input_handler.handle_all_at_once(
        ids,
        all_at_once,
        usr_levels,
        main_story.chapters,
        "treasure level",
        "chapter",
        "(&0&=none, &1&=inferior, &2&=normal, &3&=superior)",
    )
    for i, level in enumerate(usr_levels):
        if level == -1:
            continue
        if i > 2:
            i += 1
        treasure_stats[i] = [level] * 48 + [0]

    save_stats["treasures"] = treasure_stats
    print("Successfully set treasures")
    return save_stats



"""Handler for parsing csv files"""

from typing import Union
from . import helper


def remove_pkcs7_padding(data: bytes) -> bytes:
    """Remove PKCS#7 padding"""

    if len(data) % 16 != 0:
        raise ValueError("Invalid data length")
    padding_size = data[-1]
    if padding_size < 1 or padding_size > 16:
        raise ValueError("Invalid padding size")
    if data[-padding_size:] != bytes([padding_size] * padding_size):
        raise ValueError("Invalid padding")
    return data[:-padding_size]


def remove_comments(data: str) -> str:
    """Remove in-line comments from a file"""

    comments = ["#", "//"]
    data = data.split("\n")
    for comment in comments:
        data = [line.split(comment)[0] for line in data]
    data = "\n".join(data)
    return data


def parse_csv(
    file_path: str,
    delimiter: str = ",",
    remove_padding: bool = True,
    parse_int: bool = True,
    r_comments: bool = True,
    r_empty: bool = True,
) -> Union[list[list[str]], list[list[int]]]:
    """Parse csv file with error handling and return list of lists and remove padding"""

    data = helper.read_file_bytes(file_path)
    if remove_padding:
        data = remove_pkcs7_padding(data)

    lines = data.decode("utf-8")
    if r_comments:
        lines = remove_comments(lines)
    lines = lines.split("\n")
    lines = [line.strip().split(delimiter) for line in lines]
    if r_empty:
        lines = helper.remove_empty_items(lines)
    if parse_int:
        lines = helper.parse_int_list_list(lines)
    return lines



"""Handler for cat upgrades"""

from ... import helper, user_input_handler, csv_file_handler


def upgrade_cats(save_stats: dict) -> dict:
    """Upgrade specific cats"""

    cats = save_stats["cat_upgrades"]
    base = cats["Base"]
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "Enter cat ids (Look up cro battle cats to find ids)(You can enter &all& to get all, a range e.g &1&-&50&, or ids separate by spaces e.g &5 4 7&):"
        ),
        len(base),
    )
    return upgrade_cats_ids(save_stats, ids)


def upgrade_current_cats(save_stats: dict) -> dict:
    """Upgrade current cats"""

    cats = save_stats["cats"]
    cat_ids = []
    for i, cat_val in enumerate(cats):
        if cat_val == 1:
            cat_ids.append(i)
    save_stats = upgrade_cats_ids(save_stats, cat_ids)

    return save_stats


def get_rarities() -> list:
    """Get all cat ids of each rarity"""

    path = helper.get_file("game_data/rarity/unitbuy.csv")
    data = csv_file_handler.parse_csv(path)
    rarity_ids = helper.copy_first_n(data, 13)
    return rarity_ids


def get_rarity(rarity_ids: list) -> list:
    """Get all cat ids of a certain rarity"""

    rarities = get_rarities()
    cat_ids = []
    for rarity_id in rarity_ids:
        rarity_id = helper.check_int(rarity_id)
        if rarity_id is None:
            print("Please input a valid number")
            continue
        rarity_id -= 1
        for i, rarity_val in enumerate(rarities):
            if int(rarity_val) == rarity_id:
                cat_ids.append(i)
    return cat_ids


types = [
    "Normal",
    "Special",
    "Rare",
    "Super Rare",
    "Uber Super Rare",
    "Legend Rare",
]


def upgrade_cat_rarity(save_stats: dict) -> dict:
    """Upgrade all cats of a certain rarity"""

    ids = user_input_handler.select_options(
        options=types,
        mode="upgrade",
    )

    cat_ids = get_rarity(ids)
    save_stats = upgrade_cats_ids(save_stats, cat_ids)

    return save_stats


def upgrade_handler(data: dict, ids: list, item_name: str, save_stats: dict) -> dict:
    """Handler for cat upgrades"""

    ids = helper.check_cat_ids(ids, save_stats)

    base = data["Base"]
    plus = data["Plus"]
    individual = True
    if len(ids) > 1:
        individual = user_input_handler.colored_input(
            f"Do you want to upgrade each {item_name} individually(&1&), or all at once(&2&):"
        ) == "1"
    first = True
    base_lvl = None
    plus_lvl = None
    for cat_id in ids:
        if not individual and first:
            levels = get_plus_base(
                user_input_handler.colored_input(
                    'Enter the base level followed by a "&+&" then the plus level, e.g 5&+&12. If you want to ignore the base level do &+&12, if you want to ignore the plus level do 5&+&:\n'
                )
            )
            base_lvl = levels[0]
            plus_lvl = levels[1]
            first = False
        elif individual:
            helper.colored_text(f"The current upgrade level of id &{cat_id}& is &{base[cat_id]+1}&+&{plus[cat_id]}&")
            levels = get_plus_base(
                user_input_handler.colored_input(
                    f'Enter the base level for {item_name}: &{cat_id}& followed by a "&+&" then the plus level, e.g 5&+&12. If you want to ignore the base level do &+&12, if you want to ignore the plus level do 5&+&:\n'
                )
            )
            base_lvl = levels[0]
            plus_lvl = levels[1]
        if base_lvl is not None:
            base[cat_id] = base_lvl - 1
        if plus_lvl is not None:
            plus[cat_id] = plus_lvl
    data["Base"] = base
    data["Plus"] = plus

    return data


def set_user_popups(save_stats: dict) -> dict:
    """Set user popups, stops the user rank popups from spamming up the screen"""

    save_stats["user_rank_popups"]["Value"] = 0xFFFFFF
    return save_stats


def upgrade_cats_ids(save_stats: dict, ids: list) -> dict:
    """Upgrade cats by ids"""

    save_stats["cat_upgrades"] = upgrade_handler(
        data=save_stats["cat_upgrades"],
        ids=ids,
        item_name="cat",
        save_stats=save_stats,
    )
    save_stats = set_user_popups(save_stats)
    print("Successfully set cat levels")
    return save_stats


def get_plus_base(usr_input: str) -> tuple:
    """Get the base and plus level of an input"""

    split = usr_input.split("+")
    base = None
    plus = None
    if split[0]:
        base = helper.check_int(split[0])
    if split[1]:
        plus = helper.check_int(split[1])
    return base, plus



"""Handler for patching save data"""

import hashlib
from typing import Union
from . import helper

def get_md5_sum(save_data: bytes, game_version: str) -> str:
    """Get the md5 sum of the save data"""

    if game_version == "jp":
        game_version = ""
    salt = bytes("battlecats" + game_version, "utf-8")
    data_to_hash = salt + save_data[:-32]

    return hashlib.md5(data_to_hash).hexdigest()


def detect_game_version(save_data: bytes) -> Union[None, str]:
    """Detect the game version of the save file"""

    game_versions = ["jp", "en", "kr", "tw"]
    curr_hash = save_data[-32::]
    curr_hash_str = str(curr_hash, "utf-8")

    for game_version in game_versions:
        hash_str = get_md5_sum(save_data, game_version)
        if hash_str == curr_hash_str:
            return game_version
    return None


def patch_save_data(save_data: bytes, game_version: str) -> bytes:
    """Set the md5 sum of the save data"""

    save_hash = get_md5_sum(save_data, game_version)
    hash_bytes = bytes(save_hash, "utf-8")

    return helper.set_range(save_data, hash_bytes, len(save_data) - 32)



"""Handler for adb commands"""

import subprocess
import os
import sys
from typing import Union
from . import user_input_handler, helper


def adb_pull(game_version: str) -> subprocess.CompletedProcess:
    """Pull save data from specific version from a device using adb"""

    if game_version == "jp":
        game_version = ""
    path = f"/data/data/jp.co.ponos.battlecats{game_version}/files/SAVE_DATA"
    helper.colored_text(f"Pulling save data from &{path}", new=helper.GREEN)
    return adb_cmd_handler(f"adb pull \"{path}\" SAVE_DATA")


def adb_error_handler(error: str) -> None:
    """Handle adb errors, and display message if there is an error"""

    if "device '(null)' not found" in error or "device not found" in error:
        helper.colored_text(
            "Error: No device with an adb connection can be found, please connect one and try again.",
            base=helper.RED,
        )
        return None

    if "does not exist" in error:
        helper.colored_text(
            "Error: You don't seem to have this game version installed on this device please try again.",
            base=helper.RED,
        )
        return None

    if "'adb' is not recognized as an internal or external command" in error:
        add_adb = (
            user_input_handler.colored_input(
                "Error, adb is not in your Path environment variable. There is a tutorial in the github's readme. Would you like to try to add adb to your path now?(&y&/&n&):"
            )
            == "y"
        )
        if add_adb:
            add_to_path()
            print("Please re-run the editor to try again")
        sys.exit()

    helper.colored_text(f"Error: an unknown error has occurred: {error}")
    return None


def find_adb_path() -> Union[str, None]:
    """Find adb path automatically in common locations"""

    drive_letters = ["C", "D", "E"]
    for drive_letter in drive_letters:
        if os.path.exists(f"{drive_letter}:\\LDPlayer\\LDPlayer4.0"):
            return f"{drive_letter}:\\LDPlayer\\LDPlayer4.0"
        if os.path.exists(f"{drive_letter}:\\Program Files (x86)\\Nox\\bin"):
            return f"{drive_letter}:\\Program Files (x86)\\Nox\\bin"
        if os.path.exists(f"{drive_letter}:\\adb"):
            return f"{drive_letter}:\\adb"
    return None


def add_to_path() -> None:
    """Try to add adb to path environment variable automatically"""

    adb_path = find_adb_path()
    if not adb_path:
        adb_path = input(
            "Please enter the path to the folder than contains adb: download link here: https://dl.google.com/android/repository/platform-tools-latest-windows.zip:"
        )
        if os.path.isfile(adb_path):
            adb_path = os.path.dirname(adb_path)

    subprocess.run(f'setx PATH "{adb_path};%PATH%"', shell=True, check=True)
    subprocess.run(f"set PATH={adb_path};%PATH%", shell=True, check=True)
    print("Successfully added adb to path")


def adb_pull_handler(game_version: str = None) -> Union[None, str]:
    """Pull save data from device and return the path to the save data"""

    if not game_version:
        game_version = input("Enter your game version (en, jp, kr, tw):")
    if adb_pull(game_version) is None:
        return None
    return "SAVE_DATA"


def start_adb() -> None:
    """Start adb server"""

    helper.colored_text("Starting adb...", helper.GREEN)
    subprocess.run(
        "adb devices",
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        check=True,
    )


def adb_clear(game_version: str) -> None:
    """Delete save data and other required files from device for a fresh install without having to re-download any assets"""

    if game_version == "jp":
        game_version = ""
    package_name = f"jp.co.ponos.battlecats{game_version}"
    path = f"/data/data/{package_name}"

    if adb_cmd_handler(f"adb shell rm {path}/shared_prefs -r -f") is None:
        return
    if adb_cmd_handler(f"adb shell rm {path}/files/*SAVE_DATA*") is None:
        return
    adb_rerun(package_name)


def adb_rerun(package_name: str):
    """Close and open the game"""

    print("Re-opening game...")
    if adb_cmd_handler(f"adb shell am force-stop {package_name}") is None:
        return
    if adb_cmd_handler(f"adb shell monkey -p {package_name} -v 1") is None:
        return


def adb_cmd_handler(cmd: str) -> Union[subprocess.CompletedProcess, None]:
    """Handle adb commands"""

    start_adb()
    try:
        return subprocess.run(cmd, shell=True, check=True, capture_output=True)
    except subprocess.CalledProcessError as exc:
        adb_error_handler(exc.stderr.decode("utf-8"))
        return None


def adb_push(game_version: str, save_data_path: str, rerun: bool):
    """Push save data to a specific game version on device using adb"""

    version = game_version
    if version == "jp":
        version = ""
    package_name = f"jp.co.ponos.battlecats{version}"
    path = f"/data/data/{package_name}/files/SAVE_DATA"
    helper.colored_text(f"Pushing save data to &{path}&", new=helper.GREEN)

    if adb_cmd_handler(f'adb push "{save_data_path}" "{path}"') is None:
        return

    if rerun:
        adb_rerun(package_name)



"""Handler for editing the ototo cat cannon"""

from ... import user_input_handler, item

types = [
    "Base",
    "Slow Beam",
    "Iron Wall",
    "Thunderbolt",
    "Waterblast",
    "Holy Blast",
    "Breakerblast",
    "Curseblast",
]


def set_level(save_stats: dict, levels: list) -> dict:
    """Set the upgrade level of the cannon"""

    cannons = save_stats["ototo_cannon"]

    ot_levels = item.create_item_group(
        names=types,
        values=levels,
        maxes=[19, 29, 29, 29, 29, 29, 29, 29],
        edit_name="level",
        group_name="Cannon Level",
        offset=1,
    )
    ot_levels.edit()

    for i, level in enumerate(ot_levels.values):
        if level > 0:
            cannons[i]["unlock_flag"] = 3
        cannons[i]["level"] = level
    save_stats["ototo_cannon"] = cannons
    return save_stats


def set_stage(save_stats: dict, stages: list) -> dict:
    """Set the stage of the cannon development"""

    cannons = save_stats["ototo_cannon"]

    ot_stages = item.create_item_group(
        names=types[1:],
        values=stages[1:],
        maxes=3,
        edit_name="stage",
        group_name="Ototo Cat Cannon Stage",
    )
    ot_stages.edit()
    for i in range(len(cannons) - 1):
        cannons[i + 1]["unlock_flag"] = ot_stages.values[i]

    save_stats["ototo_cannon"] = cannons
    return save_stats


def edit_cat_cannon(save_stats: dict) -> dict:
    """Handler for ototo cat cannon upgrades"""

    cannons = save_stats["ototo_cannon"]
    levels = []
    stages = []
    for cannon in cannons:
        level = cannons[cannon]["level"]
        stage = cannons[cannon]["unlock_flag"]
        levels.append(level)
        stages.append(stage)
    stage = (
        user_input_handler.colored_input(
            "Do you want to set the level of the cannon &(1)& or the level of construction &(2)& (e.g foundation, style, cannon):"
        )
        == "2"
    )
    if stage:
        save_stats = set_stage(save_stats, stages)
    else:
        save_stats = set_level(save_stats, levels)
    return save_stats



"""Handler for editing gamatoto xp"""

from ... import helper, user_input_handler, item


def get_boundaries() -> list:
    """Get the xp requirements for each level"""

    boundaries = helper.read_file_string(
        helper.get_file("game_data/gamatoto/GamatotoExpedition.csv")
    ).splitlines()
    previous = 0
    xp_requirements = []
    previous = 0
    for line in boundaries:
        requirement = int(line.split(",")[0])
        if previous >= requirement:
            break
        xp_requirements.append(requirement)
        previous = requirement
    return xp_requirements


def get_level_from_xp(gamatoto_xp: int) -> dict:
    """Get the level from the xp amount"""

    xp_requirements = get_boundaries()
    level = 1
    for requirement in xp_requirements:
        if gamatoto_xp >= requirement:
            level += 1
    return {
        "level": level,
        "max_level": len(xp_requirements),
        "max_xp": xp_requirements[-2],
    }


def get_xp_from_level(level: int) -> int:
    """Get the xp amount from the level"""

    xp_requirements = get_boundaries()
    if level <= 1:
        gamatoto_xp = 0
    else:
        gamatoto_xp = xp_requirements[level - 2]
    return gamatoto_xp


def edit_gamatoto_xp(save_stats: dict) -> dict:
    """Handler for gamatoto xp"""

    gamatoto_xp = save_stats["gamatoto_xp"]

    data = get_level_from_xp(gamatoto_xp["Value"])
    level = data["level"]

    helper.colored_text(f"Gamatoto xp: &{gamatoto_xp['Value']}&\nLevel: &{level}&")
    raw = user_input_handler.colored_input(
        "Do you want to edit raw xp(&1&) or the level(&2&)?:"
    )

    if raw == "1":
        gam_xp = item.Item(
            name="Gamatoto XP",
            value=gamatoto_xp["Value"],
            max_value=None,
            edit_name="value",
        )
        gam_xp.edit()
        gamatoto_xp["Value"] = gam_xp.value
    elif raw == "2":
        gam_level = item.Item(
            name="Gamatoto Level",
            value=level,
            max_value=data["max_level"],
            edit_name="level",
        )
        gam_level.edit()
        gamatoto_xp["Value"] = get_xp_from_level(gam_level.value)
    else:
        return save_stats

    save_stats["gamatoto_xp"] = gamatoto_xp
    return save_stats



"""Handler for upgrading the blue upgrades"""

from ... import helper, user_input_handler
from ..cats import upgrade_cats

types = [
    "Power",
    "Range",
    "Charge",
    "Efficiency",
    "Wallet",
    "Health",
    "Research",
    "Accounting",
    "Study",
    "Energy",
]


def upgrade_blue_ids(save_stats: dict, ids: list) -> dict:
    """Upgrade blue upgrades for a set of ids"""

    save_stats["blue_upgrades"] = upgrade_cats.upgrade_handler(
        data=save_stats["blue_upgrades"],
        ids=ids,
        item_name="upgrade",
        save_stats=save_stats,
    )
    save_stats = upgrade_cats.set_user_popups(save_stats)
    print("Successfully set special skills")
    return save_stats


def upgrade_blue(save_stats: dict) -> dict:
    """Handler for editing blue upgrades"""

    levels = save_stats["blue_upgrades"]
    levels_removed = {
        "Base": [levels["Base"][0]] + levels["Base"][2:],
        "Plus": [levels["Plus"][0]] + levels["Plus"][2:],
    }

    levels_removed_formated = []
    for base, plus in zip(levels_removed["Base"], levels_removed["Plus"]):
        levels_removed_formated.append(
            f"{base + 1}+{plus}"
        )

    print("What do you want to upgrade:")
    helper.colored_list(types, extra_data=levels_removed_formated)

    total = len(types) + 1
    ids = user_input_handler.colored_input(
        f"{total}. &All at once&\nEnter a number from 1 to {total} (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")
    ids = user_input_handler.create_all_list(ids, 11)
    ids = helper.parse_int_list(ids, -1)
    new_ids = []
    for blue_id in ids:
        if blue_id > 0:
            blue_id += 1
        new_ids.append(blue_id)
    ids = new_ids
    save_stats = upgrade_blue_ids(save_stats, ids)
    return save_stats



"""Handler for clearing main story chapters"""

from ... import helper, user_input_handler

chapters = [
    "Empire of Cats 1",
    "Empire of Cats 2",
    "Empire of Cats 3",
    "Into the Future 1",
    "Into the Future 2",
    "Into the Future 3",
    "Cats of the Cosmos 1",
    "Cats of the Cosmos 2",
    "Cats of the Cosmos 3",
]


def specific_levels(save_stats: dict) -> dict:
    """Clear whole chapters"""

    story_chapters = save_stats["story_chapters"]

    print("What chapters do you want to select?")
    helper.colored_list(chapters)
    ids = user_input_handler.colored_input(
        "10. &All at once&\nEnter a number from 1 to 10 (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")
    ids = helper.check_clamp(ids, 10, 1, -1)
    story_chapters = specific_handler(ids, story_chapters)
    save_stats["story_chapters"] = story_chapters

    print("Successfully cleared story levels")

    return save_stats


def specific_handler(ids: list, data: dict) -> dict:
    """Clear specific stages in a chapter"""

    for chapter_id in ids:
        max_level = user_input_handler.colored_input(
            "Enter the stage id that you want to clear up to (and including) (e.g &1&=korea cleared, &2&=korea &and& mongolia cleared)?:"
        )
        max_level = helper.check_int(max_level)
        if max_level is None:
            print("Please input a number")
            return data
        max_level = helper.clamp(max_level, 1, 48)
        data["Chapter Progress"][chapter_id] = max_level
        data["Times Cleared"][chapter_id] = (
            ([1] * max_level) + ([0] * (48 - max_level)) + ([0] * 3)
        )

    return data


def main_story(save_stats: dict) -> dict:
    """Handler for editing story chapters"""

    whole = user_input_handler.colored_input(
        "Do you want to edit whole chapters at once &(1)& or specific stages &(2)&?:"
    )
    if whole == "2":
        return specific_levels(save_stats)
    story_chapters = save_stats["story_chapters"]

    print("What chapters do you want to beat completely?")
    helper.colored_list(chapters)

    ids = user_input_handler.colored_input(
        "10. &All at once&\nEnter a number from 1 to 10 (You can enter multiple values separated by spaces to edit multiple at once):"
    ).split(" ")
    ids = user_input_handler.create_all_list(ids, 10)
    ids = helper.check_clamp(ids, 10, 1, -1)

    story_chapters = edit_story(ids, story_chapters, 48)
    save_stats["story_chapters"] = story_chapters

    print("Successfully cleared story chapters")
    return save_stats


def format_story_ids(ids: list) -> list:
    """For some reason there is a gap after EoC 3. This adds that"""

    formatted_ids = []
    for story_id in ids:
        if story_id > 2:
            story_id += 1
        formatted_ids.append(story_id)
    return formatted_ids


def edit_story(ids: list, data: dict, chapter_progress: int) -> dict:
    """Edit story chapters for a set of ids"""

    ids = format_story_ids(ids)
    for chapter_id in ids:
        data["Chapter Progress"][chapter_id] = chapter_progress
        data["Times Cleared"][chapter_id] = (
            ([1] * chapter_progress) + ([0] * (48 - chapter_progress)) + ([0] * 3)
        )
    return data
